"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from pathlib import Path
import torch
from ...utils import is_mlu_available, is_musa_available, is_npu_available, is_xpu_available
from .config_args import ClusterConfig, default_json_config_file
from .config_utils import SubcommandHelpFormatter
description = "Create a default config file for SapiensAccelerator with only a few flags set."
def write_basic_config(mixed_precision="no", save_location: str = default_json_config_file, use_xpu: bool = False):
    path = Path(save_location)
    path.parent.mkdir(parents=True, exist_ok=True)
    if path.exists(): return False
    mixed_precision = mixed_precision.lower()
    if mixed_precision not in ["no", "fp16", "bf16", "fp8"]: raise ValueError(f"`mixed_precision` should be one of 'no', 'fp16', 'bf16', or 'fp8'. Received {mixed_precision}")
    config = {"compute_environment": "LOCAL_MACHINE", "mixed_precision": mixed_precision}
    if is_mlu_available():
        num_mlus = torch.mlu.device_count()
        config["num_processes"] = num_mlus
        config["use_cpu"] = False
        if num_mlus > 1: config["distributed_type"] = "MULTI_MLU"
        else: config["distributed_type"] = "NO"
    elif is_musa_available():
        num_musas = torch.musa.device_count()
        config["num_processes"] = num_musas
        config["use_cpu"] = False
        if num_musas > 1: config["distributed_type"] = "MULTI_MUSA"
        else: config["distributed_type"] = "NO"
    elif torch.cuda.is_available():
        num_gpus = torch.cuda.device_count()
        config["num_processes"] = num_gpus
        config["use_cpu"] = False
        if num_gpus > 1: config["distributed_type"] = "MULTI_GPU"
        else: config["distributed_type"] = "NO"
    elif is_xpu_available() and use_xpu:
        num_xpus = torch.xpu.device_count()
        config["num_processes"] = num_xpus
        config["use_cpu"] = False
        if num_xpus > 1: config["distributed_type"] = "MULTI_XPU"
        else: config["distributed_type"] = "NO"
    elif is_npu_available():
        num_npus = torch.npu.device_count()
        config["num_processes"] = num_npus
        config["use_cpu"] = False
        if num_npus > 1: config["distributed_type"] = "MULTI_NPU"
        else: config["distributed_type"] = "NO"
    else:
        num_xpus = 0
        config["use_cpu"] = True
        config["num_processes"] = 1
        config["distributed_type"] = "NO"
    config["debug"] = False
    config["enable_cpu_affinity"] = False
    config = ClusterConfig(**config)
    config.to_json_file(path)
    return path
def default_command_parser(parser, parents):
    parser = parser.add_parser("default", parents=parents, help=description, formatter_class=SubcommandHelpFormatter)
    parser.add_argument("--config_file", default=default_json_config_file, help=("The path to use to store the config file. Will default to a file named default_config.yaml in the cache location, which is the content of the environment `HF_HOME` suffixed with 'sapiens_accelerator', or if you don't have such an environment variable, your cache directory ('~/.cache' or the content of `XDG_CACHE_HOME`) suffixed with 'huggingface'."), dest="save_location")
    parser.add_argument("--mixed_precision", choices=["no", "fp16", "bf16"], type=str, help="Whether or not to use mixed precision training. Choose between FP16 and BF16 (bfloat16) training. BF16 training is only supported on Nvidia Ampere GPUs and PyTorch 1.10 or later.", default="no")
    parser.set_defaults(func=default_config_command)
    return parser
def default_config_command(args): pass
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import functools
from typing import Dict, List, Mapping, Optional, Union
import torch
import torch.nn as nn
from .state import PartialState
from .utils import (PrefixedDataset, find_device, named_module_tensors, send_to_device, set_module_tensor_to_device)
from .utils.memory import clear_device_cache
from .utils.modeling import get_non_persistent_buffers
from .utils.other import recursive_getattr
_sapiens_accelerator_added_attributes = ["to", "cuda", "npu", "xpu", "mlu", "musa"]
class ModelHook:
    no_grad = False
    def init_hook(self, module): return module
    def pre_forward(self, module, *args, **kwargs): return args, kwargs
    def post_forward(self, module, output): return output
    def detach_hook(self, module): return module
class SequentialHook(ModelHook):
    def __init__(self, *hooks): self.hooks = hooks
    def init_hook(self, module):
        for hook in self.hooks: module = hook.init_hook(module)
        return module
    def pre_forward(self, module, *args, **kwargs):
        for hook in self.hooks: args, kwargs = hook.pre_forward(module, *args, **kwargs)
        return args, kwargs
    def post_forward(self, module, output):
        for hook in self.hooks: output = hook.post_forward(module, output)
        return output
    def detach_hook(self, module):
        for hook in self.hooks: module = hook.detach_hook(module)
        return module
def add_hook_to_module(module: nn.Module, hook: ModelHook, append: bool = False):
    if append and (getattr(module, "_hf_hook", None) is not None):
        old_hook = module._hf_hook
        remove_hook_from_module(module)
        hook = SequentialHook(old_hook, hook)
    if hasattr(module, "_hf_hook") and hasattr(module, "_old_forward"): old_forward = module._old_forward
    else:
        old_forward = module.forward
        module._old_forward = old_forward
    module = hook.init_hook(module)
    module._hf_hook = hook
    def new_forward(module, *args, **kwargs):
        args, kwargs = module._hf_hook.pre_forward(module, *args, **kwargs)
        if module._hf_hook.no_grad:
            with torch.no_grad(): output = module._old_forward(*args, **kwargs)
        else: output = module._old_forward(*args, **kwargs)
        return module._hf_hook.post_forward(module, output)
    if "GraphModuleImpl" in str(type(module)): module.__class__.forward = functools.update_wrapper(functools.partial(new_forward, module), old_forward)
    else: module.forward = functools.update_wrapper(functools.partial(new_forward, module), old_forward)
    return module
def remove_hook_from_module(module: nn.Module, recurse=False):
    if hasattr(module, "_hf_hook"):
        module._hf_hook.detach_hook(module)
        delattr(module, "_hf_hook")
    if hasattr(module, "_old_forward"):
        if "GraphModuleImpl" in str(type(module)): module.__class__.forward = module._old_forward
        else: module.forward = module._old_forward
        delattr(module, "_old_forward")
    for attr in _sapiens_accelerator_added_attributes: module.__dict__.pop(attr, None)
    if recurse:
        for child in module.children(): remove_hook_from_module(child, recurse)
    return module
class AlignDevicesHook(ModelHook):
    def __init__(self, execution_device: Optional[Union[int, str, torch.device]] = None, offload: bool = False, io_same_device: bool = False,
    weights_map: Optional[Mapping] = None, offload_buffers: bool = False, place_submodules: bool = False, skip_keys: Optional[Union[str, List[str]]] = None,
    tied_params_map: Optional[Dict[int, Dict[torch.device, torch.Tensor]]] = None):
        self.execution_device = execution_device
        self.offload = offload
        self.io_same_device = io_same_device
        self.weights_map = weights_map
        self.offload_buffers = offload_buffers
        self.place_submodules = place_submodules
        self.skip_keys = skip_keys
        self.input_device = None
        self.param_original_devices = {}
        self.buffer_original_devices = {}
        self.tied_params_names = set()
        self.tied_params_map = tied_params_map
    def __repr__(self): return (f"AlignDevicesHook(execution_device={self.execution_device}, offload={self.offload}, io_same_device={self.io_same_device}, offload_buffers={self.offload_buffers}, place_submodules={self.place_submodules}, skip_keys={repr(self.skip_keys)})")
    def init_hook(self, module):
        if self.execution_device == "meta" or self.execution_device == torch.device("meta"): self.tied_params_map = None
        if not self.offload and self.execution_device is not None:
            for name, _ in named_module_tensors(module, recurse=self.place_submodules): set_module_tensor_to_device(module, name, self.execution_device, tied_params_map=self.tied_params_map)
        elif self.offload:
            self.original_devices = {name: param.device for name, param in named_module_tensors(module, recurse=self.place_submodules)}
            if self.weights_map is None: self.weights_map = {name: param.to("cpu") for name, param in named_module_tensors(module, include_buffers=self.offload_buffers, recurse=self.place_submodules)}
            for name, _ in named_module_tensors(module, include_buffers=self.offload_buffers, recurse=self.place_submodules, remove_non_persistent=True):
                if (self.tied_params_map is not None and recursive_getattr(module, name).data_ptr() in self.tied_params_map): self.tied_params_names.add(name)
                set_module_tensor_to_device(module, name, "meta")
            if not self.offload_buffers and self.execution_device is not None:
                for name, _ in module.named_buffers(recurse=self.place_submodules): set_module_tensor_to_device(module, name, self.execution_device, tied_params_map=self.tied_params_map)
            elif self.offload_buffers and self.execution_device is not None:
                for name in get_non_persistent_buffers(module, recurse=self.place_submodules): set_module_tensor_to_device(module, name, self.execution_device, tied_params_map=self.tied_params_map)
        return module
    def pre_forward(self, module, *args, **kwargs):
        if self.io_same_device: self.input_device = find_device([args, kwargs])
        if self.offload:
            self.tied_pointers_to_remove = set()
            for name, _ in named_module_tensors(module, include_buffers=self.offload_buffers, recurse=self.place_submodules, remove_non_persistent=True):
                fp16_statistics = None
                value = self.weights_map[name]
                if "weight" in name and name.replace("weight", "SCB") in self.weights_map.keys():
                    if value.dtype == torch.int8: fp16_statistics = self.weights_map[name.replace("weight", "SCB")]
                if name in self.tied_params_names and value.data_ptr() not in self.tied_params_map: self.tied_params_map[value.data_ptr()] = {}
                if (value is not None and self.tied_params_map is not None and value.data_ptr() in self.tied_params_map and self.execution_device not in self.tied_params_map[value.data_ptr()]): self.tied_pointers_to_remove.add((value.data_ptr(), self.execution_device))
                set_module_tensor_to_device(module, name, self.execution_device, value=value, fp16_statistics=fp16_statistics, tied_params_map=self.tied_params_map)
        return send_to_device(args, self.execution_device), send_to_device(kwargs, self.execution_device, skip_keys=self.skip_keys)
    def post_forward(self, module, output):
        if self.offload:
            for name, _ in named_module_tensors(module, include_buffers=self.offload_buffers, recurse=self.place_submodules, remove_non_persistent=True):
                set_module_tensor_to_device(module, name, "meta")
                if type(module).__name__ == "Linear8bitLt":
                    module.state.SCB = None
                    module.state.CxB = None
            for value_pointer, device in self.tied_pointers_to_remove: del self.tied_params_map[value_pointer][device]
            self.tied_pointers_to_remove = set()
        if self.io_same_device and self.input_device is not None: output = send_to_device(output, self.input_device, skip_keys=self.skip_keys)
        return output
    def detach_hook(self, module):
        if self.offload:
            for name, device in self.original_devices.items():
                if device != torch.device("meta"): set_module_tensor_to_device(module, name, device, value=self.weights_map.get(name, None))
        return module
def attach_execution_device_hook(module: torch.nn.Module, execution_device: Union[int, str, torch.device], skip_keys: Optional[Union[str, List[str]]] = None,
preload_module_classes: Optional[List[str]] = None, tied_params_map: Optional[Dict[int, Dict[torch.device, torch.Tensor]]] = None):
    if not hasattr(module, "_hf_hook") and len(module.state_dict()) > 0: add_hook_to_module(module, AlignDevicesHook(execution_device, skip_keys=skip_keys, tied_params_map=tied_params_map))
    if preload_module_classes is not None and module.__class__.__name__ in preload_module_classes: return
    for child in module.children(): attach_execution_device_hook(child, execution_device, skip_keys=skip_keys, tied_params_map=tied_params_map)
def attach_align_device_hook(module: torch.nn.Module, execution_device: Optional[torch.device] = None, offload: bool = False, weights_map: Optional[Mapping] = None,
offload_buffers: bool = False, module_name: str = "", skip_keys: Optional[Union[str, List[str]]] = None, preload_module_classes: Optional[List[str]] = None,
tied_params_map: Optional[Dict[int, Dict[torch.device, torch.Tensor]]] = None):
    directs = named_module_tensors(module)
    full_offload = (offload and preload_module_classes is not None and module.__class__.__name__ in preload_module_classes)
    if len(list(directs)) > 0 or full_offload:
        if weights_map is not None:
            prefix = f"{module_name}." if len(module_name) > 0 else ""
            prefixed_weights_map = PrefixedDataset(weights_map, prefix)
        else: prefixed_weights_map = None
        hook = AlignDevicesHook(execution_device=execution_device, offload=offload, weights_map=prefixed_weights_map, offload_buffers=offload_buffers,
        place_submodules=full_offload, skip_keys=skip_keys, tied_params_map=tied_params_map)
        add_hook_to_module(module, hook, append=True)
    if full_offload: return
    for child_name, child in module.named_children():
        child_name = f"{module_name}.{child_name}" if len(module_name) > 0 else child_name
        attach_align_device_hook(child, execution_device=execution_device, offload=offload, weights_map=weights_map, offload_buffers=offload_buffers,
        module_name=child_name, preload_module_classes=preload_module_classes, skip_keys=skip_keys, tied_params_map=tied_params_map)
def remove_hook_from_submodules(module: nn.Module):
    remove_hook_from_module(module)
    for child in module.children(): remove_hook_from_submodules(child)
def attach_align_device_hook_on_blocks(module: nn.Module, execution_device: Optional[Union[torch.device, Dict[str, torch.device]]] = None,
offload: Union[bool, Dict[str, bool]] = False, weights_map: Mapping = None, offload_buffers: bool = False, module_name: str = "", skip_keys: Optional[Union[str, List[str]]] = None,
preload_module_classes: Optional[List[str]] = None, tied_params_map: Optional[Dict[int, Dict[torch.device, torch.Tensor]]] = None):
    if not isinstance(execution_device, Mapping) and not isinstance(offload, dict):
        if not offload:
            hook = AlignDevicesHook(execution_device=execution_device, io_same_device=True, skip_keys=skip_keys, place_submodules=True, tied_params_map=tied_params_map)
            add_hook_to_module(module, hook)
        else: attach_align_device_hook(module, execution_device=execution_device, offload=True, weights_map=weights_map, offload_buffers=offload_buffers,
        module_name=module_name, skip_keys=skip_keys, tied_params_map=tied_params_map)
        return
    if not isinstance(execution_device, Mapping): execution_device = {key: execution_device for key in offload.keys()}
    if not isinstance(offload, Mapping): offload = {key: offload for key in execution_device.keys()}
    if module_name in execution_device and module_name in offload and not offload[module_name]:
        hook = AlignDevicesHook(execution_device=execution_device[module_name], offload_buffers=offload_buffers, io_same_device=(module_name == ""),
        place_submodules=True, skip_keys=skip_keys, tied_params_map=tied_params_map)
        add_hook_to_module(module, hook)
        attach_execution_device_hook(module, execution_device[module_name], skip_keys=skip_keys, tied_params_map=tied_params_map)
    elif module_name in execution_device and module_name in offload:
        attach_align_device_hook(module, execution_device=execution_device[module_name], offload=True, weights_map=weights_map, offload_buffers=offload_buffers,
        module_name=module_name, skip_keys=skip_keys, preload_module_classes=preload_module_classes, tied_params_map=tied_params_map)
        if not hasattr(module, "_hf_hook"):
            hook = AlignDevicesHook(execution_device=execution_device[module_name], io_same_device=(module_name == ""), skip_keys=skip_keys, tied_params_map=tied_params_map)
            add_hook_to_module(module, hook)
        attach_execution_device_hook(module, execution_device[module_name], preload_module_classes=preload_module_classes, skip_keys=skip_keys, tied_params_map=tied_params_map)
    elif module_name == "":
        hook = AlignDevicesHook(execution_device=execution_device.get(""), io_same_device=True, skip_keys=skip_keys, tied_params_map=tied_params_map)
        add_hook_to_module(module, hook)
    for child_name, child in module.named_children():
        child_name = f"{module_name}.{child_name}" if len(module_name) > 0 else child_name
        attach_align_device_hook_on_blocks(child, execution_device=execution_device, offload=offload, weights_map=weights_map, offload_buffers=offload_buffers,
        module_name=child_name, preload_module_classes=preload_module_classes, skip_keys=skip_keys, tied_params_map=tied_params_map)
class CpuOffload(ModelHook):
    def __init__(self, execution_device: Optional[Union[str, int, torch.device]] = None, prev_module_hook: Optional["UserCpuOffloadHook"] = None):
        self.prev_module_hook = prev_module_hook
        self.execution_device = execution_device if execution_device is not None else PartialState().default_device
    def init_hook(self, module): return module.to("cpu")
    def pre_forward(self, module, *args, **kwargs):
        if self.prev_module_hook is not None:
            self.prev_module_hook.offload()
            clear_device_cache()
        module.to(self.execution_device)
        return send_to_device(args, self.execution_device), send_to_device(kwargs, self.execution_device)
class UserCpuOffloadHook:
    def __init__(self, model, hook):
        self.model = model
        self.hook = hook
    def offload(self): self.hook.init_hook(self.model)
    def remove(self): remove_hook_from_module(self.model)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import torch
from sapiens_accelerator import Accelerator, DistributedType
class LocalSGD:
    def __enter__(self):
        if self.enabled:
            self.model_sync_obj = self.model.no_sync()
            self.model_sync_obj.__enter__()
        return self
    def __exit__(self, type, value, tb):
        if self.enabled:
            self._sync_and_avg_model_params()
            self.model_sync_obj.__exit__(type, value, tb)
    def __init__(self, accelerator: Accelerator, model: torch.nn.Module, local_sgd_steps: int, enabled: bool = True):
        if accelerator.distributed_type not in [DistributedType.NO, DistributedType.MULTI_CPU, DistributedType.MULTI_GPU, DistributedType.MULTI_MLU,
        DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU]: raise NotImplementedError("LocalSGD is supported only for CPUs and GPUs (no DeepSpeed or MegatronLM)")
        self.enabled = enabled and accelerator.distributed_type != DistributedType.NO
        self.num_steps = 0
        if self.enabled:
            self.accelerator = accelerator
            self.model = model
            self.local_sgd_steps = local_sgd_steps
    def step(self):
        self.num_steps += 1
        if not self.enabled: return
        if self.num_steps % self.local_sgd_steps == 0: self._sync_and_avg_model_params()
    def _sync_and_avg_model_params(self):
        self.accelerator.wait_for_everyone()
        with self.accelerator.autocast():
            for param in self.model.parameters(): param.data = self.accelerator.reduce(param.data, reduction="mean")
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import enum
import shutil
import sys
TERMINAL_WIDTH, _ = shutil.get_terminal_size()
CURSOR_TO_CHAR = {"UP": "A", "DOWN": "B", "RIGHT": "C", "LEFT": "D"}
class Direction(enum.Enum):
    UP = 0
    DOWN = 1
def forceWrite(content, end=""):
    sys.stdout.write(str(content) + end)
    sys.stdout.flush()
def writeColor(content, color, end=""): forceWrite(f"\u001b[{color}m{content}\u001b[0m", end)
def reset_cursor(): forceWrite("\r")
def move_cursor(num_lines: int, direction: str): forceWrite(f"\033[{num_lines}{CURSOR_TO_CHAR[direction.upper()]}")
def clear_line():
    forceWrite(" " * TERMINAL_WIDTH)
    reset_cursor()
def linebreak():
    reset_cursor()
    forceWrite("-" * TERMINAL_WIDTH)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
	########################################################################################################################################################
	# This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
	# or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
	########################################################################################################################################################
"""
import operator as op
SCALER_NAME = "scaler.pt"
MODEL_NAME = "pytorch_model"
SAFE_MODEL_NAME = "model"
RNG_STATE_NAME = "random_states"
OPTIMIZER_NAME = "optimizer"
SCHEDULER_NAME = "scheduler"
SAMPLER_NAME = "sampler"
PROFILE_PATTERN_NAME = "profile_{suffix}.json"
WEIGHTS_NAME = f"{MODEL_NAME}.bin"
WEIGHTS_PATTERN_NAME = "pytorch_model{suffix}.bin"
WEIGHTS_INDEX_NAME = f"{WEIGHTS_NAME}.index.json"
SAFE_WEIGHTS_NAME = f"{SAFE_MODEL_NAME}.safetensors"
SAFE_WEIGHTS_PATTERN_NAME = "model{suffix}.safetensors"
SAFE_WEIGHTS_INDEX_NAME = f"{SAFE_WEIGHTS_NAME}.index.json"
SAGEMAKER_PYTORCH_VERSION = "1.10.2"
SAGEMAKER_PYTHON_VERSION = "py38"
SAGEMAKER_TRANSFORMERS_VERSION = "4.17.0"
SAGEMAKER_PARALLEL_EC2_INSTANCES = ["ml.p3.16xlarge", "ml.p3dn.24xlarge", "ml.p4dn.24xlarge"]
FSDP_SHARDING_STRATEGY = ["FULL_SHARD", "SHARD_GRAD_OP", "NO_SHARD", "HYBRID_SHARD", "HYBRID_SHARD_ZERO2"]
FSDP_AUTO_WRAP_POLICY = ["TRANSFORMER_BASED_WRAP", "SIZE_BASED_WRAP", "NO_WRAP"]
FSDP_BACKWARD_PREFETCH = ["BACKWARD_PRE", "BACKWARD_POST", "NO_PREFETCH"]
FSDP_STATE_DICT_TYPE = ["FULL_STATE_DICT", "LOCAL_STATE_DICT", "SHARDED_STATE_DICT"]
FSDP_PYTORCH_VERSION = ("2.1.0.a0+32f93b1")
FSDP_MODEL_NAME = "pytorch_model_fsdp"
DEEPSPEED_MULTINODE_LAUNCHERS = ["pdsh", "standard", "openmpi", "mvapich", "mpich"]
TORCH_DYNAMO_MODES = ["default", "reduce-overhead", "max-autotune"]
ELASTIC_LOG_LINE_PREFIX_TEMPLATE_PYTORCH_VERSION = "2.2.0"
XPU_PROFILING_AVAILABLE_PYTORCH_VERSION = "2.4.0"
MITA_PROFILING_AVAILABLE_PYTORCH_VERSION = "2.1.0"
STR_OPERATION_TO_FUNC = {">": op.gt, ">=": op.ge, "==": op.eq, "!=": op.ne, "<=": op.le, "<": op.lt}
TORCH_LAUNCH_PARAMS = ["nnodes", "nproc_per_node", "rdzv_backend", "rdzv_endpoint", "rdzv_id", "rdzv_conf", "standalone", "max_restarts", "monitor_interval",
"start_method", "role", "module", "m", "no_python", "run_path", "log_dir", "r", "redirects", "t", "tee", "node_rank", "master_addr", "master_port"]
CUDA_DISTRIBUTED_TYPES = ["DEEPSPEED", "MULTI_GPU", "FSDP", "MEGATRON_LM"]
TORCH_DISTRIBUTED_OPERATION_TYPES = CUDA_DISTRIBUTED_TYPES + ["MULTI_NPU", "MULTI_MLU", "MULTI_MUSA", "MULTI_XPU", "MULTI_CPU"]
"""
	########################################################################################################################################################
	# This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
	# or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
	########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
from pathlib import Path
import evaluate
import torch
from datasets import load_dataset
from torch.optim import AdamW
from torch.utils.data import DataLoader
from transformers import AutoModelForSequenceClassification, AutoTokenizer, get_linear_schedule_with_warmup
from sapiens_accelerator import Accelerator, DeepSpeedPlugin, DistributedType
from sapiens_accelerator.state import AcceleratorState
from sapiens_accelerator.utils.deepspeed import get_active_deepspeed_plugin
MAX_GPU_BATCH_SIZE = 16
EVAL_BATCH_SIZE = 32
class NoiseModel(torch.nn.Module):
    def __init__(self, noise_factor=0.1):
        super().__init__()
        self.noise_factor = torch.nn.Parameter(torch.tensor(noise_factor, dtype=torch.float32))
    def forward(self, loss): return loss * self.noise_factor
def get_dataloaders(accelerator: Accelerator, batch_size: int = 16, model_name: str = "bert-base-cased"):
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    datasets = load_dataset("glue", "mrpc")
    def tokenize_function(examples):
        outputs = tokenizer(examples["sentence1"], examples["sentence2"], truncation=True, max_length=None)
        return outputs
    tokenized_datasets = datasets.map(tokenize_function, batched=True, remove_columns=["idx", "sentence1", "sentence2"], load_from_cache_file=False)
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")
    def collate_fn(examples):
        if accelerator.distributed_type == DistributedType.XLA: return tokenizer.pad(examples, padding="max_length", max_length=128, return_tensors="pt")
        return tokenizer.pad(examples, padding="longest", return_tensors="pt")
    train_dataloader = DataLoader(tokenized_datasets["train"], shuffle=True, collate_fn=collate_fn, batch_size=batch_size)
    eval_dataloader = DataLoader(tokenized_datasets["validation"], shuffle=False, collate_fn=collate_fn, batch_size=EVAL_BATCH_SIZE)
    return train_dataloader, eval_dataloader
test_file_path = __file__
path = Path(test_file_path).resolve()
test_file_dir_str = str(path.parent.parent.parent.parent.parent.parent)
ds_config_file = dict(zero2=f"{test_file_dir_str}/tests/deepspeed/ds_config_zero2_model_only.json", zero3=f"{test_file_dir_str}/tests/deepspeed/ds_config_zero3_model_only.json")
def single_model_training(config, args):
    num_epochs = config["num_epochs"]
    zero2_plugin = DeepSpeedPlugin(hf_ds_config=ds_config_file["zero2"])
    zero3_plugin = DeepSpeedPlugin(hf_ds_config=ds_config_file["zero3"])
    deepspeed_plugins = {"training": zero2_plugin, "inference": zero3_plugin}
    accelerator = Accelerator(deepspeed_plugins=deepspeed_plugins, mixed_precision="bf16")
    assert get_active_deepspeed_plugin(accelerator.state) is zero2_plugin
    train_model = AutoModelForSequenceClassification.from_pretrained(args.model_name_or_path)
    train_dataloader, eval_dataloader = get_dataloaders(accelerator, batch_size=config["batch_size"], model_name=args.model_name_or_path)
    max_training_steps = len(train_dataloader) * config["num_epochs"]
    optimizer = AdamW(train_model.parameters(), lr=config["lr"])
    lr_scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps=0, num_training_steps=max_training_steps)
    train_dataloader, eval_dataloader, train_model, optimizer, lr_scheduler = accelerator.prepare(train_dataloader, eval_dataloader, train_model, optimizer, lr_scheduler)
    accelerator.state.select_deepspeed_plugin("inference")
    assert get_active_deepspeed_plugin(accelerator.state) is zero3_plugin
    inference_model = NoiseModel()
    inference_model = accelerator.prepare(inference_model)
    inference_model.eval()
    accelerator.state.select_deepspeed_plugin("training")
    starting_epoch = 0
    best_performance = 0
    metric = evaluate.load("glue", "mrpc")
    performance_metric = {}
    for epoch in range(starting_epoch, num_epochs):
        train_model.train()
        inference_model.train()
        for step, batch in enumerate(train_dataloader):
            with accelerator.accumulate(train_model):
                outputs_1 = train_model(**batch)
                with torch.no_grad(): outputs_2 = inference_model(outputs_1.loss)
                loss = outputs_1.loss + outputs_2
                accelerator.backward(loss)
                optimizer.step()
                lr_scheduler.step()
                optimizer.zero_grad()
        train_model.eval()
        for step, batch in enumerate(eval_dataloader):
            with torch.no_grad(): outputs = train_model(**batch)
            predictions = outputs.logits.argmax(dim=-1)
            predictions, references = accelerator.gather_for_metrics((predictions, batch["labels"]))
            metric.add_batch(predictions=predictions, references=references)
        eval_metric = metric.compute()
        accelerator.print(f"epoch {epoch}:", eval_metric)
        performance_metric[f"epoch-{epoch}"] = eval_metric["accuracy"]
        if best_performance < eval_metric["accuracy"]: best_performance = eval_metric["accuracy"]
    assert best_performance > performance_metric["epoch-0"]
def multiple_model_training(config, args):
    num_epochs = config["num_epochs"]
    zero2_plugin = DeepSpeedPlugin(hf_ds_config=ds_config_file["zero2"])
    zero3_plugin = DeepSpeedPlugin(hf_ds_config=ds_config_file["zero3"])
    deepspeed_plugins = {"zero2": zero2_plugin, "zero3": zero3_plugin}
    zero2_accelerator = Accelerator(deepspeed_plugins=deepspeed_plugins, mixed_precision="bf16")
    zero3_accelerator = Accelerator()
    assert get_active_deepspeed_plugin(zero2_accelerator.state) is zero2_plugin
    zero2_model = AutoModelForSequenceClassification.from_pretrained(args.model_name_or_path)
    train_dataloader, eval_dataloader = get_dataloaders(zero2_accelerator, batch_size=config["batch_size"], model_name=args.model_name_or_path)
    max_training_steps = len(train_dataloader) * config["num_epochs"]
    zero2_optimizer = AdamW(zero2_model.parameters(), lr=config["lr"])
    zero2_lr_scheduler = get_linear_schedule_with_warmup(zero2_optimizer, num_warmup_steps=0, num_training_steps=max_training_steps)
    train_dataloader, eval_dataloader, zero2_model, zero2_optimizer, zero2_lr_scheduler = zero2_accelerator.prepare(train_dataloader, eval_dataloader, zero2_model, zero2_optimizer, zero2_lr_scheduler)
    assert zero2_accelerator.deepspeed_engine_wrapped.engine is zero2_model
    zero3_accelerator.state.select_deepspeed_plugin("zero3")
    zero3_plugin.deepspeed_config["train_micro_batch_size_per_gpu"] = zero2_plugin.deepspeed_config["train_micro_batch_size_per_gpu"]
    assert get_active_deepspeed_plugin(zero3_accelerator.state) is zero3_plugin
    zero3_model = AutoModelForSequenceClassification.from_pretrained(args.model_name_or_path)
    zero3_optimizer = AdamW(zero3_model.parameters(), lr=config["lr"])
    zero3_lr_scheduler = get_linear_schedule_with_warmup(zero3_optimizer, num_warmup_steps=0, num_training_steps=max_training_steps)
    zero3_model, zero3_optimizer, zero3_lr_scheduler = zero3_accelerator.prepare(zero3_model, zero3_optimizer, zero3_lr_scheduler)
    assert zero3_accelerator.deepspeed_engine_wrapped.engine is zero3_model
    starting_epoch = 0
    best_performance_a = 0
    best_performance_b = 0
    metric_a = evaluate.load("glue", "mrpc")
    metric_b = evaluate.load("glue", "mrpc")
    performance_metric_a = {}
    performance_metric_b = {}
    for epoch in range(starting_epoch, num_epochs):
        zero2_model.train()
        zero3_model.train()
        for step, batch in enumerate(train_dataloader):
            with zero2_accelerator.accumulate(zero2_model, zero3_model):
                outputs_1 = zero2_model(**batch)
                zero2_accelerator.backward(outputs_1.loss)
                zero2_optimizer.step()
                zero2_lr_scheduler.step()
                zero2_optimizer.zero_grad()
                outputs_2 = zero3_model(**batch)
                zero3_accelerator.backward(outputs_2.loss)
                zero3_optimizer.step()
                zero3_lr_scheduler.step()
                zero3_optimizer.zero_grad()
        zero2_model.eval()
        zero3_model.eval()
        for step, batch in enumerate(eval_dataloader):
            with torch.no_grad():
                logits_a = zero2_model(**batch).logits
                logits_b = zero3_model(**batch).logits
            predictions_a = logits_a.argmax(dim=-1)
            predictions_b = logits_b.argmax(dim=-1)
            predictions_a, predictions_b, references = zero2_accelerator.gather_for_metrics((predictions_a, predictions_b, batch["labels"]))
            metric_a.add_batch(predictions=predictions_a, references=references)
            metric_b.add_batch(predictions=predictions_b, references=references)
        eval_metric_a = metric_a.compute()
        eval_metric_b = metric_b.compute()
        zero2_accelerator.print(f"epoch {epoch}:", eval_metric_a, eval_metric_b)
        performance_metric_a[f"epoch-{epoch}"] = eval_metric_a["accuracy"]
        performance_metric_b[f"epoch-{epoch}"] = eval_metric_b["accuracy"]
        if best_performance_a < eval_metric_a["accuracy"]: best_performance_a = eval_metric_a["accuracy"]
        if best_performance_b < eval_metric_b["accuracy"]: best_performance_b = eval_metric_b["accuracy"]
    assert best_performance_a > performance_metric_a["epoch-0"]
    assert best_performance_b > performance_metric_b["epoch-0"]
def main():
    parser = argparse.ArgumentParser(description="Simple example of training script tracking peak GPU memory usage.")
    parser.add_argument("--model_name_or_path", type=str, default="bert-base-cased", help="", required=False)
    parser.add_argument("--performance_lower_bound", type=float, default=None, help="Optional lower bound for the performance metric. If set, the training will throw error when the performance metric drops below this value.")
    parser.add_argument("--num_epochs", type=int, default=2, help="Number of train epochs.")
    args = parser.parse_args()
    config = {"lr": 2e-5, "num_epochs": args.num_epochs, "seed": 42, "batch_size": 16}
    single_model_training(config, args)
    AcceleratorState._reset_state(True)
    multiple_model_training(config, args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from __future__ import annotations
import logging
import os
import threading
import warnings
from contextlib import contextmanager
from functools import partial
from typing import Any, Callable, Optional
import torch
from .utils import (DistributedType, DynamoBackend, GradientAccumulationPlugin, check_cuda_p2p_ib_support, check_fp8_capability, deepspeed_required,
get_ccl_version, get_cpu_distributed_information, get_int_from_env, is_ccl_available, is_datasets_available, is_deepspeed_available, is_fp8_available,
is_ipex_available, is_mlu_available, is_mps_available, is_musa_available, is_npu_available, is_torch_xla_available, is_xpu_available, parse_choice_from_env,
parse_flag_from_env, set_numa_affinity)
from .utils.dataclasses import SageMakerDistributedType
if is_torch_xla_available(): import torch_xla.core.xla_model as xm
if is_mlu_available(check_device=False): import torch_mlu
if is_musa_available(check_device=False): import torch_musa
if is_npu_available(check_device=False): import torch_npu
logger = logging.getLogger(__name__)
def is_initialized() -> bool: return AcceleratorState._shared_state != {}
def do_nothing(*args, **kwargs): return None
class ThreadLocalSharedDict(threading.local):
    def __init__(self, thread_local: bool = False): self._storage = {}
    def __get__(self, obj, objtype=None): return self._storage
    def __set__(self, obj, value): self._storage = value
SharedDict = dict if not is_torch_xla_available() else ThreadLocalSharedDict
class PartialState:
    _shared_state = SharedDict()
    _known_attrs = ["_cpu", "_mixed_precision", "_shared_state", "backend", "debug", "device", "distributed_type", "fork_launched", "local_process_index", "num_processes", "process_index"]
    def __init__(self, cpu: bool = False, **kwargs):
        self.__dict__ = self._shared_state
        if not self.initialized:
            self._cpu = cpu
            self.backend = None
            env_device = os.environ.get("SAPIENS_ACCELERATOR_TORCH_DEVICE", None)
            self.device = torch.device(env_device) if env_device is not None else None
            self.debug = parse_flag_from_env("SAPIENS_ACCELERATOR_DEBUG_MODE")
            use_sagemaker_dp = kwargs.pop("_use_sagemaker_dp", None)
            dist_information = None
            if use_sagemaker_dp is None: use_sagemaker_dp = (os.environ.get("SAPIENS_ACCELERATOR_USE_SAGEMAKER", "false") == "true" and os.environ.get("SAPIENS_ACCELERATOR_SAGEMAKER_DISTRIBUTED_TYPE") != SageMakerDistributedType.NO)
            original_backend = kwargs.pop("backend", None)
            backend, distributed_type = self._prepare_backend(cpu, use_sagemaker_dp, original_backend)
            if original_backend is not None and backend != original_backend: raise ValueError(f"Your assigned backend {original_backend} is not avaliable, please use {backend}")
            self.backend = backend
            self.distributed_type = distributed_type
            use_deepspeed = False
            if not cpu and self.backend != "xla":
                if int(os.environ.get("LOCAL_RANK", -1)) != -1:
                    if os.environ.get("SAPIENS_ACCELERATOR_USE_DEEPSPEED", "false") == "true":
                        if not is_deepspeed_available(): raise ImportError("DeepSpeed is not available => install it using `pip3 install deepspeed` or build it from source")
                        from deepspeed import comm as dist
                        if not dist.is_initialized(): dist.init_distributed(dist_backend=self.backend, auto_mpi_discovery=False, **kwargs)
                        use_deepspeed = True
                    elif (self.distributed_type not in (DistributedType.MULTI_XPU, DistributedType.MULTI_CPU) and not torch.distributed.is_initialized()): torch.distributed.init_process_group(backend=self.backend, **kwargs)
            if self.distributed_type in (DistributedType.MULTI_XPU, DistributedType.MULTI_CPU):
                dist_information = get_cpu_distributed_information()
                os.environ["RANK"] = str(dist_information.rank)
                os.environ["WORLD_SIZE"] = str(dist_information.world_size)
                os.environ["LOCAL_RANK"] = str(dist_information.local_rank)
                os.environ["LOCAL_WORLD_SIZE"] = str(dist_information.local_world_size)
                if not os.environ.get("MASTER_PORT", None): os.environ["MASTER_PORT"] = "29500"
                if (not os.environ.get("MASTER_ADDR", None) and dist_information.local_world_size != dist_information.world_size and self.backend != "mpi"): raise ValueError("Tried to launch on distributed with multinode, but `MASTER_ADDR` env was not set, please try exporting rank 0's hostname as `MASTER_ADDR`")
                kwargs["rank"] = dist_information.rank
                kwargs["world_size"] = dist_information.world_size
                if (self.distributed_type == DistributedType.MULTI_CPU and get_int_from_env(["OMP_NUM_THREADS"], 0) == 0):
                    import psutil
                    num_cpu_threads_per_process = int(psutil.cpu_count(logical=False) / dist_information.local_world_size)
                    if num_cpu_threads_per_process == 0: num_cpu_threads_per_process = 1
                    torch.set_num_threads(num_cpu_threads_per_process)
                if not torch.distributed.is_initialized(): torch.distributed.init_process_group(backend=self.backend, **kwargs)
            if self.backend is None:
                self.distributed_type = DistributedType.NO
                self.num_processes = 1
                self.process_index = 0
                self.local_process_index = 0
            elif self.backend == "xla":
                self.set_device()
                xm.set_replication(self.device, xm.get_xla_supported_devices())
                self.num_processes = xm.xrt_world_size()
                self.process_index = xm.get_ordinal()
                if is_torch_xla_available(check_is_tpu=True): self.local_process_index = xm.get_local_ordinal()
                else: self.local_process_index = int(os.environ.get("LOCAL_RANK", -1))
            else:
                self.num_processes = torch.distributed.get_world_size()
                self.process_index = torch.distributed.get_rank()
                self.local_process_index = (int(os.environ.get("LOCAL_RANK", -1)) if dist_information is None else dist_information.local_rank)
            self.set_device()
            if use_deepspeed: self.distributed_type = DistributedType.DEEPSPEED
            if parse_flag_from_env("SAPIENS_ACCELERATOR_CPU_AFFINITY", False): set_numa_affinity(self.local_process_index)
            if self.device.type == "cuda" and not check_cuda_p2p_ib_support():
                if "NCCL_P2P_DISABLE" not in os.environ or "NCCL_IB_DISABLE" not in os.environ: raise NotImplementedError("Using RTX 4000 series doesn't support faster communication broadband via P2P or IB. "+'Please set `NCCL_P2P_DISABLE="1"` and `NCCL_IB_DISABLE="1" or use `sapiens_accelerator launch` which '+"will do this automatically.")
        self.fork_launched = parse_flag_from_env("FORK_LAUNCHED", 0)
    def __repr__(self) -> str: return (f"Distributed environment: {self.distributed_type}{('  Backend: ' + self.backend) if self.backend else ''}\nNum processes: {self.num_processes}\nProcess index: {self.process_index}\nLocal process index: {self.local_process_index}\nDevice: {self.device}\n")
    @staticmethod
    def _reset_state(): PartialState._shared_state.clear()
    @property
    def initialized(self) -> bool: return self._shared_state != {}
    @property
    def use_distributed(self): return self.distributed_type != DistributedType.NO and self.num_processes > 1
    @property
    def is_last_process(self) -> bool: return self.process_index == self.num_processes - 1
    @property
    def is_main_process(self) -> bool: return (self.process_index == 0 if self.distributed_type != DistributedType.MEGATRON_LM else self.is_last_process)
    @property
    def is_local_main_process(self) -> bool: return (self.local_process_index == 0 if self.distributed_type != DistributedType.MEGATRON_LM else self.is_last_process)
    def wait_for_everyone(self):
        if self.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU,
        DistributedType.MULTI_XPU, DistributedType.MULTI_CPU, DistributedType.DEEPSPEED, DistributedType.FSDP): torch.distributed.barrier()
        elif self.distributed_type == DistributedType.XLA: xm.rendezvous("sapiens_accelerator.utils.wait_for_everyone")
    def _goes_first(self, is_main: bool):
        if not is_main: self.wait_for_everyone()
        yield
        if is_main: self.wait_for_everyone()
    @contextmanager
    def split_between_processes(self, inputs: list | tuple | dict | torch.Tensor, apply_padding: bool = False):
        if self.num_processes == 1:
            yield inputs
            return
        length = len(inputs)
        if isinstance(inputs, dict):
            length = len(inputs[list(inputs.keys())[0]])
            if not all(len(v) == length for v in inputs.values()): raise ValueError("All values in the dictionary must have the same length")
        num_samples_per_process, num_extras = divmod(length, self.num_processes)
        start_index = self.process_index * num_samples_per_process + min(self.process_index, num_extras)
        end_index = start_index + num_samples_per_process + (1 if self.process_index < num_extras else 0)
        def _split_values(inputs, start_index, end_index):
            if isinstance(inputs, (list, tuple, torch.Tensor)):
                if start_index >= len(inputs): result = inputs[-1:]
                else: result = inputs[start_index:end_index]
                if apply_padding:
                    if isinstance(result, torch.Tensor):
                        from sapiens_accelerator.utils import pad_across_processes, send_to_device
                        tensorized_result = send_to_device(result, self.device)
                        result = pad_across_processes(tensorized_result, pad_index=inputs[-1])
                    else: result += [result[-1]] * (num_samples_per_process + 1 - len(result))
                return result
            elif isinstance(inputs, dict):
                for key in inputs.keys(): inputs[key] = _split_values(inputs[key], start_index, end_index)
                return inputs
            else:
                if is_datasets_available():
                    from datasets import Dataset
                    if isinstance(inputs, Dataset):
                        if start_index >= len(inputs): start_index = len(inputs) - 1
                        if end_index > len(inputs): end_index = len(inputs)
                        result_idcs = list(range(start_index, end_index))
                        if apply_padding: result_idcs += [end_index - 1] * (num_samples_per_process + 1 - len(result_idcs))
                        return inputs.select(result_idcs)
                return inputs
        yield _split_values(inputs, start_index, end_index)
    @contextmanager
    def main_process_first(self): yield from self._goes_first(self.is_main_process)
    @contextmanager
    def local_main_process_first(self): yield from self._goes_first(self.is_local_main_process)
    def on_main_process(self, function: Callable[..., Any] = None):
        if not self.initialized: raise ValueError("The `PartialState` or `Accelerator` must be initialized before calling this function.")
        if self.is_main_process or not self.use_distributed: return function
        return do_nothing
    def on_local_main_process(self, function: Callable[..., Any] = None):
        if self.is_local_main_process or not self.use_distributed: return function
        return do_nothing
    def on_last_process(self, function: Callable[..., Any]):
        if self.is_last_process or not self.use_distributed: return function
        return do_nothing
    def on_process(self, function: Callable[..., Any] = None, process_index: int = None):
        if function is None: return partial(self.on_process, process_index=process_index)
        if (self.process_index == process_index) or (not self.use_distributed): return function
        return do_nothing
    def on_local_process(self, function: Callable[..., Any] = None, local_process_index: int = None):
        if function is None: return partial(self.on_local_process, local_process_index=local_process_index)
        if (self.local_process_index == local_process_index) or (not self.use_distributed): return function
        return do_nothing
    def print(self, *args, **kwargs): pass
    @property
    def default_device(self) -> torch.device:
        if is_mps_available():
            os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
            return torch.device("mps")
        elif is_mlu_available(): return torch.device("mlu")
        elif is_musa_available(): return torch.device("musa")
        elif is_npu_available(): return torch.device("npu")
        elif torch.cuda.is_available(): return torch.device("cuda")
        elif is_xpu_available(): return torch.device("xpu")
        else: return torch.device("cpu")
    def _prepare_backend(self, cpu: bool = False, sagemaker_dp=False, backend: str = None) -> tuple[str, DistributedType]:
        distributed_type = None
        if sagemaker_dp:
            import smdistributed.dataparallel.torch.torch_smddp
            backend = "smddp"
            distributed_type = DistributedType.MULTI_GPU
        elif is_torch_xla_available():
            backend = "xla"
            distributed_type = DistributedType.XLA
        elif int(os.environ.get("LOCAL_RANK", -1)) != -1 and not cpu:
            if is_mlu_available():
                backend = "cncl"
                distributed_type = DistributedType.MULTI_MLU
            elif is_musa_available():
                backend = "mccl"
                distributed_type = DistributedType.MULTI_MUSA
            elif is_npu_available():
                backend = "hccl"
                distributed_type = DistributedType.MULTI_NPU
            elif torch.cuda.is_available():
                if backend is None: backend = "nccl"
                distributed_type = DistributedType.MULTI_GPU
        if distributed_type is None and (int(os.environ.get("LOCAL_RANK", -1)) != -1 or get_int_from_env(["PMI_SIZE", "OMPI_COMM_WORLD_SIZE", "MV2_COMM_WORLD_SIZE", "WORLD_SIZE"], 1) > 1):
            if not cpu and is_xpu_available(): distributed_type = DistributedType.MULTI_XPU
            else: distributed_type = DistributedType.MULTI_CPU
            if (backend in (None, "ccl") and is_ccl_available() and (get_int_from_env(["CCL_WORKER_COUNT"], 0) > 0 or distributed_type == DistributedType.MULTI_XPU)):
                if get_ccl_version() >= "1.12": import oneccl_bindings_for_pytorch
                else: import torch_ccl
                backend = "ccl"
            elif backend in (None, "mpi") and torch.distributed.is_mpi_available(): backend = "mpi"
            else: backend = "gloo"
        if distributed_type is None: distributed_type = DistributedType.NO
        return backend, distributed_type
    def set_device(self):
        if self.device is not None: return
        if self.distributed_type == DistributedType.NO:
            self.device = torch.device("cpu") if self._cpu else self.default_device
            return
        device = str(self.distributed_type).split(".")[-1].replace("MULTI_", "").lower()
        if device not in ("cpu", "gpu", "mlu", "musa", "npu", "xpu", "xla"): raise ValueError(f"Can't set device for {self.distributed_type} ({device}), verify we should be calling `_set_device()` for it!")
        if device == "xla": self.device = xm.xla_device()
        else:
            if device == "gpu": device = "cuda"
            device_module = getattr(torch, device)
            device_index = self.local_process_index % device_module.device_count()
            self.device = torch.device(device, device_index)
            device_module.set_device(self.device)
    def destroy_process_group(self, group=None):
        if self.fork_launched and group is None: return
        if torch.distributed.is_initialized(): torch.distributed.destroy_process_group(group)
    def __getattr__(self, name: str):
        if name in self._known_attrs: raise AttributeError(f"`PartialState` object has no attribute `{name}`. This happens if `PartialState._reset_state()` was called and an `Accelerator` or `PartialState` was not reinitialized.")
        raise AttributeError(f"'PartialState' object has no attribute '{name}'")
class AcceleratorState:
    _shared_state = SharedDict()
    _known_attrs = PartialState._known_attrs + ["deepspeed_plugin", "use_ipex", "fsdp_plugin", "megatron_lm_plugin", "dynamo_plugin"]
    def __init__(self, mixed_precision: str = None, cpu: bool = False, dynamo_plugin=None, deepspeed_plugin=None, fsdp_plugin=None, megatron_lm_plugin=None,
    _from_accelerator: bool = False, **kwargs):
        self.__dict__ = self._shared_state
        if parse_flag_from_env("SAPIENS_ACCELERATOR_USE_CPU"): cpu = True
        if PartialState._shared_state == {}: PartialState(cpu, **kwargs)
        self.__dict__.update(PartialState._shared_state)
        self._check_initialized(mixed_precision, cpu)
        if not self.initialized:
            self.deepspeed_plugins = None
            self.use_ipex = None
            mixed_precision = (parse_choice_from_env("SAPIENS_ACCELERATOR_MIXED_PRECISION", "no") if mixed_precision is None else mixed_precision.lower())
            if mixed_precision == "fp8":
                if not is_fp8_available(): raise ValueError("Using `fp8` precision requires `transformer_engine` or `MS-AMP` to be installed.")
                elif not check_fp8_capability(): mixed_precision = "fp16"
            self.dynamo_plugin = dynamo_plugin
            if not _from_accelerator: raise ValueError("Please make sure to properly initialize your accelerator via `accelerator = Accelerator()` before using any functionality from the `sapiens_accelerator` library.")
            self._mixed_precision = "no" if self.distributed_type == DistributedType.DEEPSPEED else mixed_precision
            if self.distributed_type == DistributedType.XLA and is_torch_xla_available(check_is_tpu=True):
                if mixed_precision == "bf16":
                    if os.environ.get("SAPIENS_ACCELERATOR_DOWNCAST_BF16"):
                        os.environ["XLA_USE_BF16"] = str(0)
                        os.environ["XLA_DOWNCAST_BF16"] = str(1)
                        self.downcast_bfloat = True
                    else:
                        os.environ["XLA_USE_BF16"] = str(1)
                        os.environ["XLA_DOWNCAST_BF16"] = str(0)
                        self.downcast_bfloat = False
            elif os.environ.get("SAPIENS_ACCELERATOR_USE_DEEPSPEED", "false") == "true" and not cpu:
                self.deepspeed_plugins = deepspeed_plugin
                self.distributed_type = DistributedType.DEEPSPEED
            elif self.distributed_type in [DistributedType.MULTI_GPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU, DistributedType.MULTI_XPU]:
                if os.environ.get("SAPIENS_ACCELERATOR_USE_FSDP", "false") == "true" or fsdp_plugin is not None:
                    self.distributed_type = DistributedType.FSDP
                    if self._mixed_precision != "no": fsdp_plugin.set_mixed_precision(self._mixed_precision)
                    self.fsdp_plugin = fsdp_plugin
                if os.environ.get("SAPIENS_ACCELERATOR_USE_MEGATRON_LM", "false") == "true" and self.distributed_type not in [DistributedType.MULTI_XPU]:
                    self.distributed_type = DistributedType.MEGATRON_LM
                    megatron_lm_plugin.set_mixed_precision(self._mixed_precision)
                    self.megatron_lm_plugin = megatron_lm_plugin
            elif self.distributed_type in [DistributedType.MULTI_CPU, DistributedType.MULTI_XPU, DistributedType.NO]:
                if is_ipex_available(): self.use_ipex = parse_flag_from_env("SAPIENS_ACCELERATOR_USE_IPEX", default=True)
                else: self.use_ipex = False
            if (self.dynamo_plugin.backend != DynamoBackend.NO and self._mixed_precision == "no" and self.device.type == "cuda"): torch.backends.cuda.matmul.allow_tf32 = True
            if (self.dynamo_plugin.backend != DynamoBackend.NO and self._mixed_precision == "no" and self.device.type == "musa"): torch.backends.musa.matmul.allow_tf32 = True
            PartialState._shared_state["distributed_type"] = self.distributed_type
    @property
    def initialized(self) -> bool: return self._shared_state != PartialState._shared_state
    def __repr__(self):
        repr = PartialState().__repr__() + f"\nMixed precision type: {self.mixed_precision}\n"
        if self.distributed_type == DistributedType.DEEPSPEED: repr += f"ds_config: {self.deepspeed_plugin.deepspeed_config}\n"
        return repr
    def _check_initialized(self, mixed_precision=None, cpu=None):
        if self.initialized:
            err = "AcceleratorState has already been initialized and cannot be changed, restart your runtime completely and pass `{flag}` to `Accelerator()`."
            if cpu and self.device.type != "cpu": raise ValueError(err.format(flag="cpu=True"))
            if (mixed_precision is not None and mixed_precision != self._mixed_precision and self.distributed_type != DistributedType.DEEPSPEED): raise ValueError(err.format(flag=f"mixed_precision='{mixed_precision}'"))
    @property
    def mixed_precision(self):
        if self.distributed_type == DistributedType.DEEPSPEED:
            config = self.deepspeed_plugin.deepspeed_config
            if config.get("fp16", {}).get("enabled", False): mixed_precision = "fp16"
            elif config.get("bf16", {}).get("enabled", False): mixed_precision = "bf16"
            else: mixed_precision = "no"
        else: mixed_precision = self._mixed_precision
        return mixed_precision
    @staticmethod
    def _reset_state(reset_partial_state: bool = False):
        AcceleratorState._shared_state.clear()
        if reset_partial_state: PartialState._reset_state()
    def destroy_process_group(self, group=None): PartialState().destroy_process_group(group)
    @property
    def fork_launched(self): return PartialState().fork_launched
    @property
    def use_distributed(self): return PartialState().use_distributed
    @property
    def is_last_process(self) -> bool: return PartialState().is_last_process
    @property
    def is_main_process(self) -> bool: return PartialState().is_main_process
    @property
    def is_local_main_process(self) -> bool: return PartialState().is_local_main_process
    def wait_for_everyone(self): PartialState().wait_for_everyone()
    @contextmanager
    def split_between_processes(self, inputs: list | tuple | dict | torch.Tensor, apply_padding: bool = False):
        with PartialState().split_between_processes(inputs, apply_padding=apply_padding) as inputs: yield inputs
    @contextmanager
    def main_process_first(self):
        with PartialState().main_process_first(): yield
    @contextmanager
    def local_main_process_first(self):
        with PartialState().local_main_process_first(): yield
    @property
    def deepspeed_plugin(self):
        if self.distributed_type != DistributedType.DEEPSPEED: return None
        from sapiens_accelerator.utils.deepspeed import get_active_deepspeed_plugin
        return get_active_deepspeed_plugin(self)
    @deepspeed_required
    def get_deepspeed_plugin(self, name: str): return self.deepspeed_plugins[name]
    @deepspeed_required
    def select_deepspeed_plugin(self, name: str = None):
        for key, plugin in self.deepspeed_plugins.items():
            if key != name: plugin._unselect()
        self.deepspeed_plugins[name].select(_from_accelerator_state=True)
    def print(self, *args, **kwargs): PartialState().print(*args, **kwargs)
    def __getattr__(self, name: str):
        if name in self._known_attrs: raise AttributeError(f"`AcceleratorState` object has no attribute `{name}`. This happens if `AcceleratorState._reset_state()` was called and an `Accelerator` or `PartialState` was not reinitialized.")
        raise AttributeError(f"'AcceleratorState' object has no attribute '{name}'")
class GradientState:
    _shared_state = SharedDict()
    def __init__(self, gradient_accumulation_plugin: Optional[GradientAccumulationPlugin] = None):
        self.__dict__ = self._shared_state
        if not self.initialized:
            self.sync_gradients = True
            self.active_dataloader = None
            self.dataloader_references = [None]
            self.plugin_kwargs = (gradient_accumulation_plugin.to_kwargs() if gradient_accumulation_plugin is not None else {})
            self._is_xla_gradients_synced = False
        if gradient_accumulation_plugin is not None and self.plugin_kwargs != gradient_accumulation_plugin.to_kwargs(): self.plugin_kwargs = gradient_accumulation_plugin.to_kwargs()
    @property
    def num_steps(self) -> int: return self.plugin_kwargs.get("num_steps", 1)
    @property
    def adjust_scheduler(self) -> bool: return self.plugin_kwargs.get("adjust_scheduler", False)
    @property
    def sync_with_dataloader(self) -> bool: return self.plugin_kwargs.get("sync_with_dataloader", True)
    @property
    def initialized(self) -> bool: return GradientState._shared_state != {}
    @property
    def end_of_dataloader(self) -> bool:
        if not self.in_dataloader: return False
        return self.active_dataloader.end_of_dataloader
    @property
    def remainder(self) -> int:
        if not self.in_dataloader: return -1
        return self.active_dataloader.remainder
    def __repr__(self): return (f"Sync Gradients: {self.sync_gradients}\nAt end of current dataloader: {self.end_of_dataloader}\nExtra samples added: {self.remainder}\nGradient accumulation plugin: {self.plugin_kwargs}\n")
    @property
    def is_xla_gradients_synced(self):
        if parse_flag_from_env("SAPIENS_ACCELERATOR_USE_FSDP", default=False): return True
        return self._is_xla_gradients_synced
    @is_xla_gradients_synced.setter
    def is_xla_gradients_synced(self, is_synced): self._is_xla_gradients_synced = is_synced
    def _set_sync_gradients(self, sync_gradients):
        self.sync_gradients = sync_gradients
        if (self.sync_gradients and is_torch_xla_available(check_is_tpu=True) and PartialState().distributed_type == DistributedType.XLA): xm.mark_step()
    def _add_dataloader(self, dataloader):
        self.active_dataloader = dataloader
        self.dataloader_references.append(self.active_dataloader)
    def _remove_dataloader(self, dataloader):
        self.dataloader_references.remove(dataloader)
        self.active_dataloader = self.dataloader_references[-1]
    @property
    def in_dataloader(self) -> bool: return self.active_dataloader is not None
    @staticmethod
    def _reset_state(): GradientState._shared_state.clear()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import json
import os
from ...utils.constants import SAGEMAKER_PARALLEL_EC2_INSTANCES, TORCH_DYNAMO_MODES
from ...utils.dataclasses import ComputeEnvironment, SageMakerDistributedType
from ...utils.imports import is_boto3_available
from .config_args import SageMakerConfig
from .config_utils import (DYNAMO_BACKENDS, _ask_field, _ask_options, _convert_dynamo_backend, _convert_mixed_precision, _convert_sagemaker_distributed_mode, _convert_yes_no_to_bool)
if is_boto3_available(): import boto3
def _create_iam_role_for_sagemaker(role_name):
    iam_client = boto3.client("iam")
    sagemaker_trust_policy = {"Version": "2012-10-17", "Statement": [{"Effect": "Allow", "Principal": {"Service": "sagemaker.amazonaws.com"}, "Action": "sts:AssumeRole"}]}
    try:
        iam_client.create_role(RoleName=role_name, AssumeRolePolicyDocument=json.dumps(sagemaker_trust_policy, indent=2))
        policy_document = {"Version": "2012-10-17", "Statement": [{"Effect": "Allow", "Action": ["sagemaker:*", "ecr:GetDownloadUrlForLayer", "ecr:BatchGetImage",
        "ecr:BatchCheckLayerAvailability", "ecr:GetAuthorizationToken", "cloudwatch:PutMetricData", "cloudwatch:GetMetricData", "cloudwatch:GetMetricStatistics",
        "cloudwatch:ListMetrics", "logs:CreateLogGroup", "logs:CreateLogStream", "logs:DescribeLogStreams", "logs:PutLogEvents", "logs:GetLogEvents",
        "s3:CreateBucket", "s3:ListBucket", "s3:GetBucketLocation", "s3:GetObject", "s3:PutObject"], "Resource": "*"}]}
        iam_client.put_role_policy(RoleName=role_name, PolicyName=f"{role_name}_policy_permission", PolicyDocument=json.dumps(policy_document, indent=2))
    except iam_client.exceptions.EntityAlreadyExistsException: pass
def _get_iam_role_arn(role_name):
    iam_client = boto3.client("iam")
    return iam_client.get_role(RoleName=role_name)["Role"]["Arn"]
def get_sagemaker_input():
    credentials_configuration = _ask_options("How do you want to authorize?", ["AWS Profile", "Credentials (AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY) "], int)
    aws_profile = None
    if credentials_configuration == 0:
        aws_profile = _ask_field("Enter your AWS Profile name: [default] ", default="default")
        os.environ["AWS_PROFILE"] = aws_profile
    else:
        aws_access_key_id = _ask_field("AWS Access Key ID: ")
        os.environ["AWS_ACCESS_KEY_ID"] = aws_access_key_id
        aws_secret_access_key = _ask_field("AWS Secret Access Key: ")
        os.environ["AWS_SECRET_ACCESS_KEY"] = aws_secret_access_key
    aws_region = _ask_field("Enter your AWS Region: [us-east-1]", default="us-east-1")
    os.environ["AWS_DEFAULT_REGION"] = aws_region
    role_management = _ask_options("Do you already have an IAM Role for executing Amazon SageMaker Training Jobs?", ["Provide IAM Role name", "Create new IAM role using credentials"], int)
    if role_management == 0: iam_role_name = _ask_field("Enter your IAM role name: ")
    else:
        iam_role_name = "sapiens_accelerator_sagemaker_execution_role"
        _create_iam_role_for_sagemaker(iam_role_name)
    is_custom_docker_image = _ask_field("Do you want to use custom Docker image? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    docker_image = None
    if is_custom_docker_image: docker_image = _ask_field("Enter your Docker image: ", lambda x: str(x).lower())
    is_sagemaker_inputs_enabled = _ask_field("Do you want to provide SageMaker input channels with data locations? [yes/NO]: ", _convert_yes_no_to_bool,
    default=False, error_message="Please enter yes or no.")
    sagemaker_inputs_file = None
    if is_sagemaker_inputs_enabled: sagemaker_inputs_file = _ask_field("Enter the path to the SageMaker inputs TSV file with columns (channel_name, data_location): ", lambda x: str(x).lower())
    is_sagemaker_metrics_enabled = _ask_field("Do you want to enable SageMaker metrics? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    sagemaker_metrics_file = None
    if is_sagemaker_metrics_enabled: sagemaker_metrics_file = _ask_field("Enter the path to the SageMaker metrics TSV file with columns (metric_name, metric_regex): ", lambda x: str(x).lower())
    distributed_type = _ask_options("What is the distributed mode?", ["No distributed training", "Data parallelism"], _convert_sagemaker_distributed_mode)
    dynamo_config = {}
    use_dynamo = _ask_field("Do you wish to optimize your script with torch dynamo?[yes/NO]:", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    if use_dynamo:
        prefix = "dynamo_"
        dynamo_config[prefix + "backend"] = _ask_options("Which dynamo backend would you like to use?", [x.lower() for x in DYNAMO_BACKENDS], _convert_dynamo_backend, default=2)
        use_custom_options = _ask_field("Do you want to customize the defaults sent to torch.compile? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
        if use_custom_options:
            dynamo_config[prefix + "mode"] = _ask_options("Which mode do you want to use?", TORCH_DYNAMO_MODES, lambda x: TORCH_DYNAMO_MODES[int(x)], default="default")
            dynamo_config[prefix + "use_fullgraph"] = _ask_field("Do you want the fullgraph mode or it is ok to break model into several subgraphs? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            dynamo_config[prefix + "use_dynamic"] = _ask_field("Do you want to enable dynamic shape tracing? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    ec2_instance_query = "Which EC2 instance type you want to use for your training?"
    if distributed_type != SageMakerDistributedType.NO: ec2_instance_type = _ask_options(ec2_instance_query, SAGEMAKER_PARALLEL_EC2_INSTANCES, lambda x: SAGEMAKER_PARALLEL_EC2_INSTANCES[int(x)])
    else:
        ec2_instance_query += "? [ml.p3.2xlarge]:"
        ec2_instance_type = _ask_field(ec2_instance_query, lambda x: str(x).lower(), default="ml.p3.2xlarge")
    debug = False
    if distributed_type != SageMakerDistributedType.NO: debug = _ask_field("Should distributed operations be checked while running for errors? This can avoid timeout issues but will be slower. [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    num_machines = 1
    if distributed_type in (SageMakerDistributedType.DATA_PARALLEL, SageMakerDistributedType.MODEL_PARALLEL): num_machines = _ask_field("How many machines do you want use? [1]: ", int, default=1)
    mixed_precision = _ask_options("Do you wish to use FP16 or BF16 (mixed precision)?", ["no", "fp16", "bf16", "fp8"], _convert_mixed_precision)
    return SageMakerConfig(image_uri=docker_image, compute_environment=ComputeEnvironment.AMAZON_SAGEMAKER, distributed_type=distributed_type, use_cpu=False,
    dynamo_config=dynamo_config, ec2_instance_type=ec2_instance_type, profile=aws_profile, region=aws_region, iam_role_name=iam_role_name, mixed_precision=mixed_precision,
    num_machines=num_machines, sagemaker_inputs_file=sagemaker_inputs_file, sagemaker_metrics_file=sagemaker_metrics_file, debug=debug)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import json
import os
import time
from functools import wraps
from typing import Any, Dict, List, Optional, Union
import yaml
from .logging import get_logger
from .state import PartialState
from .utils import (LoggerType, is_aim_available, is_clearml_available, is_comet_ml_available, is_dvclive_available, is_mlflow_available,
is_tensorboard_available, is_wandb_available, listify)
_available_trackers = []
if is_tensorboard_available(): _available_trackers.append(LoggerType.TENSORBOARD)
if is_wandb_available(): _available_trackers.append(LoggerType.WANDB)
if is_comet_ml_available(): _available_trackers.append(LoggerType.COMETML)
if is_aim_available(): _available_trackers.append(LoggerType.AIM)
if is_mlflow_available(): _available_trackers.append(LoggerType.MLFLOW)
if is_clearml_available(): _available_trackers.append(LoggerType.CLEARML)
if is_dvclive_available(): _available_trackers.append(LoggerType.DVCLIVE)
logger = get_logger(__name__)
def on_main_process(function):
    @wraps(function)
    def execute_on_main_process(self, *args, **kwargs):
        if getattr(self, "main_process_only", False): return PartialState().on_main_process(function)(self, *args, **kwargs)
        else: return function(self, *args, **kwargs)
    return execute_on_main_process
def get_available_trackers(): return _available_trackers
class GeneralTracker:
    main_process_only = True
    def __init__(self, _blank=False):
        if not _blank:
            err = ""
            if not hasattr(self, "name"): err += "`name`"
            if not hasattr(self, "requires_logging_directory"):
                if len(err) > 0: err += ", "
                err += "`requires_logging_directory`"
            if "tracker" not in dir(self):
                if len(err) > 0: err += ", "
                err += "`tracker`"
            if len(err) > 0: raise NotImplementedError(f"The implementation for this tracker class is missing the following required attributes. Please define them in the class definition: {err}")
    def store_init_configuration(self, values: dict): pass
    def log(self, values: dict, step: Optional[int], **kwargs): pass
    def finish(self): pass
class TensorBoardTracker(GeneralTracker):
    name = "tensorboard"
    requires_logging_directory = True
    @on_main_process
    def __init__(self, run_name: str, logging_dir: Union[str, os.PathLike], **kwargs):
        try: from torch.utils import tensorboard
        except ModuleNotFoundError: import tensorboardX as tensorboard
        super().__init__()
        self.run_name = run_name
        self.logging_dir = os.path.join(logging_dir, run_name)
        self.writer = tensorboard.SummaryWriter(self.logging_dir, **kwargs)
    @property
    def tracker(self): return self.writer
    @on_main_process
    def store_init_configuration(self, values: dict):
        self.writer.add_hparams(values, metric_dict={})
        self.writer.flush()
        project_run_name = time.time()
        dir_name = os.path.join(self.logging_dir, str(project_run_name))
        os.makedirs(dir_name, exist_ok=True)
        with open(os.path.join(dir_name, "hparams.yml"), "w") as outfile:
            try: yaml.dump(values, outfile)
            except yaml.representer.RepresenterError: raise
    @on_main_process
    def log(self, values: dict, step: Optional[int] = None, **kwargs):
        values = listify(values)
        for k, v in values.items():
            if isinstance(v, (int, float)): self.writer.add_scalar(k, v, global_step=step, **kwargs)
            elif isinstance(v, str): self.writer.add_text(k, v, global_step=step, **kwargs)
            elif isinstance(v, dict): self.writer.add_scalars(k, v, global_step=step, **kwargs)
        self.writer.flush()
    @on_main_process
    def log_images(self, values: dict, step: Optional[int], **kwargs):
        for k, v in values.items(): self.writer.add_images(k, v, global_step=step, **kwargs)
    @on_main_process
    def finish(self): self.writer.close()
class WandBTracker(GeneralTracker):
    name = "wandb"
    requires_logging_directory = False
    main_process_only = False
    @on_main_process
    def __init__(self, run_name: str, **kwargs):
        super().__init__()
        self.run_name = run_name
        import wandb
        self.run = wandb.init(project=self.run_name, **kwargs)
    @property
    def tracker(self): return self.run
    @on_main_process
    def store_init_configuration(self, values: dict):
        import wandb
        wandb.config.update(values, allow_val_change=True)
    @on_main_process
    def log(self, values: dict, step: Optional[int] = None, **kwargs): self.run.log(values, step=step, **kwargs)
    @on_main_process
    def log_images(self, values: dict, step: Optional[int] = None, **kwargs):
        import wandb
        for k, v in values.items(): self.log({k: [wandb.Image(image) for image in v]}, step=step, **kwargs)
    @on_main_process
    def log_table(self, table_name: str, columns: List[str] = None, data: List[List[Any]] = None, dataframe: Any = None, step: Optional[int] = None, **kwargs):
        import wandb
        values = {table_name: wandb.Table(columns=columns, data=data, dataframe=dataframe)}
        self.log(values, step=step, **kwargs)
    @on_main_process
    def finish(self): self.run.finish()
class CometMLTracker(GeneralTracker):
    name = "comet_ml"
    requires_logging_directory = False
    @on_main_process
    def __init__(self, run_name: str, **kwargs):
        super().__init__()
        self.run_name = run_name
        from comet_ml import Experiment
        self.writer = Experiment(project_name=run_name, **kwargs)
    @property
    def tracker(self): return self.writer
    @on_main_process
    def store_init_configuration(self, values: dict): self.writer.log_parameters(values)
    @on_main_process
    def log(self, values: dict, step: Optional[int] = None, **kwargs):
        if step is not None: self.writer.set_step(step)
        for k, v in values.items():
            if isinstance(v, (int, float)): self.writer.log_metric(k, v, step=step, **kwargs)
            elif isinstance(v, str): self.writer.log_other(k, v, **kwargs)
            elif isinstance(v, dict): self.writer.log_metrics(v, step=step, **kwargs)
    @on_main_process
    def finish(self): self.writer.end()
class AimTracker(GeneralTracker):
    name = "aim"
    requires_logging_directory = True
    @on_main_process
    def __init__(self, run_name: str, logging_dir: Optional[Union[str, os.PathLike]] = ".", **kwargs):
        self.run_name = run_name
        from aim import Run
        self.writer = Run(repo=logging_dir, **kwargs)
        self.writer.name = self.run_name
    @property
    def tracker(self): return self.writer
    @on_main_process
    def store_init_configuration(self, values: dict): self.writer["hparams"] = values
    @on_main_process
    def log(self, values: dict, step: Optional[int], **kwargs):
        for key, value in values.items(): self.writer.track(value, name=key, step=step, **kwargs)
    @on_main_process
    def log_images(self, values: dict, step: Optional[int] = None, kwargs: Optional[Dict[str, dict]] = None):
        import aim
        aim_image_kw = {}
        track_kw = {}
        if kwargs is not None:
            aim_image_kw = kwargs.get("aim_image", {})
            track_kw = kwargs.get("track", {})
        for key, value in values.items():
            if isinstance(value, tuple): img, caption = value
            else: img, caption = value, ""
            aim_image = aim.Image(img, caption=caption, **aim_image_kw)
            self.writer.track(aim_image, name=key, step=step, **track_kw)
    @on_main_process
    def finish(self): self.writer.close()
class MLflowTracker(GeneralTracker):
    name = "mlflow"
    requires_logging_directory = False
    @on_main_process
    def __init__(self, experiment_name: str = None, logging_dir: Optional[Union[str, os.PathLike]] = None, run_id: Optional[str] = None, tags: Optional[Union[Dict[str, Any], str]] = None,
    nested_run: Optional[bool] = False, run_name: Optional[str] = None, description: Optional[str] = None):
        experiment_name = os.environ.get("MLFLOW_EXPERIMENT_NAME", experiment_name)
        run_id = os.environ.get("MLFLOW_RUN_ID", run_id)
        tags = os.environ.get("MLFLOW_TAGS", tags)
        if isinstance(tags, str): tags = json.loads(tags)
        nested_run = os.environ.get("MLFLOW_NESTED_RUN", nested_run)
        import mlflow
        exps = mlflow.search_experiments(filter_string=f"name = '{experiment_name}'")
        if len(exps) > 0: experiment_id = exps[0].experiment_id
        else: experiment_id = mlflow.create_experiment(name=experiment_name, artifact_location=logging_dir, tags=tags)
        self.active_run = mlflow.start_run(run_id=run_id, experiment_id=experiment_id, run_name=run_name, nested=nested_run, tags=tags, description=description)
    @property
    def tracker(self): return self.active_run
    @on_main_process
    def store_init_configuration(self, values: dict):
        import mlflow
        for name, value in list(values.items()):
            if len(str(value)) > mlflow.utils.validation.MAX_PARAM_VAL_LENGTH: del values[name]
        values_list = list(values.items())
        for i in range(0, len(values_list), mlflow.utils.validation.MAX_PARAMS_TAGS_PER_BATCH): mlflow.log_params(dict(values_list[i : i + mlflow.utils.validation.MAX_PARAMS_TAGS_PER_BATCH]))
    @on_main_process
    def log(self, values: dict, step: Optional[int]):
        metrics = {}
        for k, v in values.items():
            if isinstance(v, (int, float)): metrics[k] = v
        import mlflow
        mlflow.log_metrics(metrics, step=step)
    @on_main_process
    def finish(self):
        import mlflow
        mlflow.end_run()
class ClearMLTracker(GeneralTracker):
    name = "clearml"
    requires_logging_directory = False
    @on_main_process
    def __init__(self, run_name: str = None, **kwargs):
        from clearml import Task
        current_task = Task.current_task()
        self._initialized_externally = False
        if current_task:
            self._initialized_externally = True
            self.task = current_task
            return
        kwargs.setdefault("project_name", os.environ.get("CLEARML_PROJECT", run_name))
        kwargs.setdefault("task_name", os.environ.get("CLEARML_TASK", run_name))
        self.task = Task.init(**kwargs)
    @property
    def tracker(self): return self.task
    @on_main_process
    def store_init_configuration(self, values: dict): return self.task.connect_configuration(values)
    @on_main_process
    def log(self, values: Dict[str, Union[int, float]], step: Optional[int] = None, **kwargs):
        clearml_logger = self.task.get_logger()
        for k, v in values.items():
            if not isinstance(v, (int, float)): continue
            if step is None:
                clearml_logger.report_single_value(name=k, value=v, **kwargs)
                continue
            title, series = ClearMLTracker._get_title_series(k)
            clearml_logger.report_scalar(title=title, series=series, value=v, iteration=step, **kwargs)
    @on_main_process
    def log_images(self, values: dict, step: Optional[int] = None, **kwargs):
        clearml_logger = self.task.get_logger()
        for k, v in values.items():
            title, series = ClearMLTracker._get_title_series(k)
            clearml_logger.report_image(title=title, series=series, iteration=step, image=v, **kwargs)
    @on_main_process
    def log_table(self, table_name: str, columns: List[str] = None, data: List[List[Any]] = None, dataframe: Any = None, step: Optional[int] = None, **kwargs):
        to_report = dataframe
        if dataframe is None:
            if data is None: raise ValueError("`ClearMLTracker.log_table` requires that `data` to be supplied if `dataframe` is `None`")
            to_report = [columns] + data if columns else data
        title, series = ClearMLTracker._get_title_series(table_name)
        self.task.get_logger().report_table(title=title, series=series, table_plot=to_report, iteration=step, **kwargs)
    @on_main_process
    def finish(self):
        if self.task and not self._initialized_externally: self.task.close()
    @staticmethod
    def _get_title_series(name):
        for prefix in ["eval", "test", "train"]:
            if name.startswith(prefix + "_"): return name[len(prefix) + 1 :], prefix
        return name, "train"
class DVCLiveTracker(GeneralTracker):
    name = "dvclive"
    requires_logging_directory = False
    @on_main_process
    def __init__(self, run_name: Optional[str] = None, live: Optional[Any] = None, **kwargs):
        from dvclive import Live
        super().__init__()
        self.live = live if live is not None else Live(**kwargs)
    @property
    def tracker(self): return self.live
    @on_main_process
    def store_init_configuration(self, values: dict): self.live.log_params(values)
    @on_main_process
    def log(self, values: dict, step: Optional[int] = None, **kwargs):
        from dvclive.plots import Metric
        if step is not None: self.live.step = step
        for k, v in values.items():
            if Metric.could_log(v): self.live.log_metric(k, v, **kwargs)
        self.live.next_step()
    @on_main_process
    def finish(self): self.live.end()
LOGGER_TYPE_TO_CLASS = {"aim": AimTracker, "comet_ml": CometMLTracker, "mlflow": MLflowTracker, "tensorboard": TensorBoardTracker, "wandb": WandBTracker, "clearml": ClearMLTracker, "dvclive": DVCLiveTracker}
def filter_trackers(log_with: List[Union[str, LoggerType, GeneralTracker]], logging_dir: Union[str, os.PathLike] = None): return []
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import os
import sys
from contextlib import contextmanager
if os.name == "nt":
    import ctypes
    import msvcrt
    class CursorInfo(ctypes.Structure): _fields_ = [("size", ctypes.c_int), ("visible", ctypes.c_byte)]
def hide_cursor():
    if os.name == "nt":
        ci = CursorInfo()
        handle = ctypes.windll.kernel32.GetStdHandle(-11)
        ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
        ci.visible = False
        ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))
    elif os.name == "posix":
        sys.stdout.write("\033[?25l")
        sys.stdout.flush()
def show_cursor():
    if os.name == "nt":
        ci = CursorInfo()
        handle = ctypes.windll.kernel32.GetStdHandle(-11)
        ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
        ci.visible = True
        ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))
    elif os.name == "posix":
        sys.stdout.write("\033[?25h")
        sys.stdout.flush()
@contextmanager
def hide():
    try:
        hide_cursor()
        yield
    finally: show_cursor()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from sapiens_accelerator.commands.utils import CustomArgumentParser
from sapiens_accelerator.utils import merge_fsdp_weights
description = "Utility to merge the weights from multiple FSDP checkpoints into a single combined checkpoint. Should be used if `SHARDED_STATE_DICT` was used for the model. Weights will be saved to `{output_path}`. This is a CPU-bound process and requires enough RAM to load the entire model state dict."
def merge_command(args): merge_fsdp_weights(args.checkpoint_directory, args.output_path, not args.unsafe_serialization, args.remove_checkpoint_dir)
def merge_command_parser(subparsers=None):
    if subparsers is not None: parser = subparsers.add_parser("merge-weights", description=description)
    else: parser = CustomArgumentParser(description=description)
    parser.add_argument("checkpoint_directory", type=str, help="A directory containing sharded weights saved by FSDP.")
    parser.add_argument("output_path", type=str, help="The path to save the merged weights. Defaults to the current directory.")
    parser.add_argument("--unsafe_serialization", action="store_false", default=False, help="Whether to save the merged weights as `.bin` rather than `.safetensors` (not recommended).")
    parser.add_argument("--remove_checkpoint_dir", action="store_true", help="Whether to remove the checkpoint directory after merging.", default=False)
    if subparsers is not None: parser.set_defaults(func=merge_command)
    return parser
def main():
    parser = merge_command_parser()
    args = parser.parse_args()
    merge_command(args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import os
from sapiens_accelerator.utils import ComputeEnvironment
from .cluster import get_cluster_input
from .config_args import cache_dir, default_config_file, default_yaml_config_file, load_config_from_file
from .config_utils import _ask_field, _ask_options, _convert_compute_environment
from .sagemaker import get_sagemaker_input
description = "Launches a series of prompts to create and save a `default_config.yaml` configuration file for your training system. Should always be ran first on your machine"
def get_user_input():
    compute_environment = _ask_options("In which compute environment are you running?", ["This machine", "AWS (Amazon SageMaker)"], _convert_compute_environment)
    if compute_environment == ComputeEnvironment.AMAZON_SAGEMAKER: config = get_sagemaker_input()
    else: config = get_cluster_input()
    return config
def config_command_parser(subparsers=None):
    if subparsers is not None: parser = subparsers.add_parser("config", description=description)
    else: parser = argparse.ArgumentParser("SapiensAccelerator config command", description=description)
    parser.add_argument("--config_file", default=None, help=("The path to use to store the config file. Will default to a file named default_config.yaml in the cache location, which is the content of the environment `HF_HOME` suffixed with 'sapiens_accelerator', or if you don't have such an environment variable, your cache directory ('~/.cache' or the content of `XDG_CACHE_HOME`) suffixed with 'huggingface'."))
    if subparsers is not None: parser.set_defaults(func=config_command)
    return parser
def config_command(args):
    config = get_user_input()
    if args.config_file is not None: config_file = args.config_file
    else:
        if not os.path.isdir(cache_dir): os.makedirs(cache_dir)
        config_file = default_yaml_config_file
    if config_file.endswith(".json"): config.to_json_file(config_file)
    else: config.to_yaml_file(config_file)
def main():
    parser = config_command_parser()
    args = parser.parse_args()
    config_command(args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import math
import os
from abc import ABC
from functools import partial
import torch
import torch.nn.functional as F
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
from torch.nn.parallel.distributed import DistributedDataParallel as torchDDP
from ..optimizer import SapiensAcceleratordOptimizer
from ..scheduler import SapiensAcceleratordScheduler
from .imports import is_megatron_lm_available
from .operations import recursively_apply, send_to_device
if is_megatron_lm_available():
    from megatron import (get_args, get_num_microbatches, get_tensorboard_writer, get_tokenizer, print_rank_last)
    from megatron.arguments import (_add_data_args, _add_validation_args, core_transformer_config_from_args, parse_args, validate_args)
    from megatron.checkpointing import load_args_from_checkpoint, load_checkpoint, save_checkpoint
    from megatron.core import mpu, tensor_parallel
    from megatron.core.distributed import DistributedDataParallel as LocalDDP
    from megatron.core.distributed import finalize_model_grads
    from megatron.core.enums import ModelType
    from megatron.core.parallel_state import get_tensor_model_parallel_group, get_tensor_model_parallel_src_rank
    from megatron.core.pipeline_parallel import get_forward_backward_func
    from megatron.core.utils import get_model_config
    from megatron.data.dataset_utils import build_train_valid_test_datasets
    from megatron.global_vars import set_global_variables
    from megatron.initialize import (_compile_dependencies, _init_autoresume, _initialize_distributed, _set_random_seed, set_jit_fusion_options, write_args_to_tensorboard)
    from megatron.model import BertModel, Float16Module, GPTModel, T5Model
    from megatron.model.classification import Classification
    from megatron.optimizer import get_megatron_optimizer
    from megatron.text_generation.communication import broadcast_int_list, broadcast_tensor
    from megatron.text_generation.generation import (beam_search_and_return_on_first_stage, generate_tokens_probs_and_return_on_first_stage)
    from megatron.tokenizer.tokenizer import _vocab_size_with_padding
    from megatron.training import (build_train_valid_test_data_iterators, get_optimizer_param_scheduler, num_floating_point_operations, setup_model_and_optimizer, train_step, training_log)
    from megatron.utils import (average_losses_across_data_parallel_group, calc_params_l2_norm, get_ltor_masks_and_position_ids, unwrap_model)
def model_provider_func(pre_process=True, post_process=True, add_encoder=True, add_decoder=True):
    args = get_args()
    mode = "pre-training" if args.pretraining_flag else "fine-tuning"
    config = core_transformer_config_from_args(args)
    if args.model_type_name == "bert":
        if args.pretraining_flag:
            num_tokentypes = 2 if args.bert_binary_head else 0
            model = BertModel(config=config, num_tokentypes=num_tokentypes, add_binary_head=args.bert_binary_head, parallel_output=True, pre_process=pre_process, post_process=post_process)
        else: model = Classification(config=config, num_classes=args.num_labels, num_tokentypes=2, pre_process=pre_process, post_process=post_process)
    elif args.model_type_name == "gpt": model = GPTModel(config=config, num_tokentypes=0, parallel_output=True, pre_process=pre_process, post_process=post_process)
    elif args.model_type_name == "t5": model = T5Model(config=config, num_tokentypes=0, parallel_output=True, pre_process=pre_process, post_process=post_process, add_encoder=add_encoder, add_decoder=add_decoder)
    else: raise ValueError(f"Unsupported model type: {args.model_type_name}")
    return model
def prepare_model_optimizer_scheduler(accelerator):
    accelerator.print("Preparing model optimizer scheduler")
    args = get_args()
    if accelerator.state.megatron_lm_plugin.custom_prepare_model_function is not None:
        if accelerator.state.megatron_lm_plugin.custom_model_provider_function is None: raise ValueError("You must provide a `custom_model_provider_function` when using a `custom_prepare_model_function`.")
        custom_model_provider_func = accelerator.state.megatron_lm_plugin.custom_model_provider_function
        model = accelerator.state.megatron_lm_plugin.custom_prepare_model_function(custom_model_provider_func)
        optimizer = prepare_optimizer(accelerator, model)
        scheduler = prepare_scheduler(accelerator, optimizer, scheduler=None)
    else:
        model_type = ModelType.encoder_or_decoder
        if args.model_type_name == "t5": model_type = ModelType.encoder_and_decoder
        model_provider_func_ = model_provider_func
        if accelerator.state.megatron_lm_plugin.custom_model_provider_function is not None: model_provider_func_ = accelerator.state.megatron_lm_plugin.custom_model_provider_function
        (model, optimizer, scheduler) = setup_model_and_optimizer(model_provider_func_, model_type, no_wd_decay_cond=args.no_wd_decay_cond, scale_lr_cond=args.scale_lr_cond, lr_mult=args.lr_mult)
    args.model_len = len(model)
    return model, optimizer, scheduler
class MegatronLMDummyDataLoader:
    def __init__(self, **dataset_kwargs):
        parser = argparse.ArgumentParser()
        parser = _add_data_args(parser)
        parser = _add_validation_args(parser)
        data_args = parser.parse_known_args()
        self.dataset_args = vars(data_args[0])
        self.dataset_args.update(dataset_kwargs)
        self.dataset_args["megatron_dataset_flag"] = True
    def set_megatron_data_args(self):
        args = get_args()
        for key, value in self.dataset_args.items():
            old_value = getattr(args, key, "")
            setattr(args, key, value)
    def get_train_valid_test_datasets_provider(self, accelerator):
        def train_valid_test_datasets_provider(train_val_test_num_samples):
            args = get_args()
            dataset_args = {"data_prefix": args.data_path if isinstance(args.data_path, (list, tuple)) else [args.data_path], "splits_string": args.split, "train_valid_test_num_samples": train_val_test_num_samples, "seed": args.seed}
            if args.model_type_name == "bert": dataset_args.update({"max_seq_length": args.seq_length, "binary_head": args.bert_binary_head})
            elif args.model_type_name == "gpt": dataset_args.update({"max_seq_length": args.seq_length})
            elif args.model_type_name == "t5": dataset_args.update({"max_seq_length": args.encoder_seq_length, "max_seq_length_dec": args.decoder_seq_length, "dataset_type": "t5"})
            else: raise ValueError(f"Unsupported model type: {args.model_type_name}")
            train_ds, valid_ds, test_ds = build_train_valid_test_datasets(**dataset_args)
            return train_ds, valid_ds, test_ds
        if accelerator.state.megatron_lm_plugin.custom_megatron_datasets_provider_function is not None: return accelerator.state.megatron_lm_plugin.custom_megatron_datasets_provider_function
        try:
            args = get_args()
            if args.model_type_name == "bert":
                from pretrain_bert import train_valid_test_datasets_provider
                train_valid_test_datasets_provider.is_distributed = True
                return train_valid_test_datasets_provider
            elif args.model_type_name == "gpt":
                from pretrain_gpt import train_valid_test_datasets_provider
                train_valid_test_datasets_provider.is_distributed = True
                return train_valid_test_datasets_provider
            elif args.model_type_name == "t5":
                from pretrain_t5 import train_valid_test_datasets_provider
                train_valid_test_datasets_provider.is_distributed = True
                return train_valid_test_datasets_provider
        except ImportError: pass
        return train_valid_test_datasets_provider
    def build_train_valid_test_data_iterators(self, accelerator):
        args = get_args()
        train_valid_test_dataset_provider = self.get_train_valid_test_datasets_provider(accelerator)
        if args.virtual_pipeline_model_parallel_size is not None:
            train_data_iterator = []
            valid_data_iterator = []
            test_data_iterator = []
            for i in range(getattr(args, "model_len", 0)):
                mpu.set_virtual_pipeline_model_parallel_rank(i)
                iterators = build_train_valid_test_data_iterators(train_valid_test_dataset_provider)
                train_data_iterator.append(iterators[0])
                valid_data_iterator.append(iterators[1])
                test_data_iterator.append(iterators[2])
        else: train_data_iterator, valid_data_iterator, test_data_iterator = build_train_valid_test_data_iterators(train_valid_test_dataset_provider)
        return train_data_iterator, valid_data_iterator, test_data_iterator
def _handle_megatron_data_iterator(accelerator, data_iterator):
    class DummyMegatronDataloader:
        def __iter__(self): return self
        def __next__(self): return {}
    is_data_iterator_empty = data_iterator is None
    is_src_data_iterator_empty = torch.tensor(is_data_iterator_empty, dtype=torch.bool, device=accelerator.device)
    torch.distributed.broadcast(is_src_data_iterator_empty, get_tensor_model_parallel_src_rank(), group=get_tensor_model_parallel_group())
    if not is_src_data_iterator_empty and is_data_iterator_empty: return DummyMegatronDataloader()
    return data_iterator
def prepare_data_loader(accelerator, dataloader):
    accelerator.print("Preparing dataloader")
    args = get_args()
    if not args.megatron_dataset_flag:
        from ..data_loader import _PYTORCH_DATALOADER_KWARGS, prepare_data_loader
        micro_batch_size = args.micro_batch_size * args.num_micro_batches
        kwargs = {k: getattr(dataloader, k, _PYTORCH_DATALOADER_KWARGS[k]) for k in _PYTORCH_DATALOADER_KWARGS}
        if kwargs["batch_size"] is None:
            if isinstance(kwargs["sampler"], torch.utils.data.BatchSampler): kwargs["sampler"].batch_size = micro_batch_size
            else:
                del kwargs["sampler"]
                del kwargs["shuffle"]
                del kwargs["batch_size"]
                kwargs["batch_sampler"].batch_size = micro_batch_size
        else:
            del kwargs["batch_sampler"]
            kwargs["batch_size"] = micro_batch_size
        dataloader = torch.utils.data.DataLoader(dataloader.dataset, **kwargs)
        return prepare_data_loader(dataloader, accelerator.device, num_processes=mpu.get_data_parallel_world_size(), process_index=mpu.get_data_parallel_rank(),
        split_batches=False, put_on_device=True, rng_types=accelerator.rng_types.copy(), dispatch_batches=accelerator.dispatch_batches)
    else:
        if args.consumed_samples is not None: (args.consumed_train_samples, args.consumed_valid_samples, args.consumed_test_samples) = args.consumed_samples
        else: args.consumed_train_samples, args.consumed_valid_samples, args.consumed_test_samples = 0, 0, 0
        args.micro_batch_size = args.micro_batch_size * args.num_micro_batches
        (train_data_iterator, valid_data_iterator, test_data_iterator) = dataloader.build_train_valid_test_data_iterators(accelerator)
        args.micro_batch_size = args.micro_batch_size // args.num_micro_batches
        train_data_iterator = _handle_megatron_data_iterator(accelerator=accelerator, data_iterator=train_data_iterator)
        valid_data_iterator = _handle_megatron_data_iterator(accelerator=accelerator, data_iterator=valid_data_iterator)
        test_data_iterator = _handle_megatron_data_iterator(accelerator=accelerator, data_iterator=test_data_iterator)
        return train_data_iterator, valid_data_iterator, test_data_iterator
class MegatronLMOptimizerWrapper(SapiensAcceleratordOptimizer):
    def __init__(self, optimizer): super().__init__(optimizer, device_placement=False, scaler=None)
    def zero_grad(self, set_to_none=None): pass
    def step(self): pass
    @property
    def step_was_skipped(self): return self.optimizer.skipped_iter
def prepare_optimizer(accelerator, model):
    accelerator.print("Preparing optimizer")
    args = get_args()
    return get_megatron_optimizer(model, args.no_wd_decay_cond, args.scale_lr_cond, args.lr_mult)
class MegatronLMDummyScheduler:
    def __init__(self, optimizer, total_num_steps=None, warmup_num_steps=0, **kwargs):
        self.optimizer = optimizer
        self.total_num_steps = total_num_steps
        self.warmup_num_steps = warmup_num_steps
        self.kwargs = kwargs
class MegatronLMSchedulerWrapper(SapiensAcceleratordScheduler):
    def __init__(self, scheduler, optimizers): super().__init__(scheduler, optimizers)
    def step(self, *args, **kwargs): return
def prepare_scheduler(accelerator, optimizer, scheduler):
    accelerator.print("Preparing scheduler")
    scheduler = get_optimizer_param_scheduler(optimizer)
    return scheduler
class AbstractTrainStep(ABC):
    def __init__(self, name):
        super().__init__()
        self.name = name
    def get_batch_func(self, accelerator, megatron_dataset_flag): pass
    def get_forward_step_func(self): pass
    def get_loss_func(self, accelerator): pass
class BertTrainStep(AbstractTrainStep):
    def __init__(self, accelerator, args):
        super().__init__("BertTrainStep")
        self.get_batch = self.get_batch_func(accelerator, args.megatron_dataset_flag)
        self.loss_func = self.get_loss_func(accelerator, args.pretraining_flag, args.num_labels)
        self.forward_step = self.get_forward_step_func(args.pretraining_flag, args.bert_binary_head)
        if not args.model_return_dict: self.model_output_class = None
        else:
            from transformers.modeling_outputs import SequenceClassifierOutput
            self.model_output_class = SequenceClassifierOutput
    def get_batch_func(self, accelerator, megatron_dataset_flag):
        def get_batch_megatron(data_iterator):
            keys = ["text", "types", "labels", "is_random", "loss_mask", "padding_mask"]
            datatype = torch.int64
            if data_iterator is not None: data = next(data_iterator)
            else: data = None
            data_b = tensor_parallel.broadcast_data(keys, data, datatype)
            tokens = data_b["text"].long()
            types = data_b["types"].long()
            sentence_order = data_b["is_random"].long()
            loss_mask = data_b["loss_mask"].float()
            lm_labels = data_b["labels"].long()
            padding_mask = data_b["padding_mask"].long()
            return tokens, types, sentence_order, loss_mask, lm_labels, padding_mask
        def get_batch_transformer(data_iterator):
            data = next(data_iterator)
            data = send_to_device(data, torch.cuda.current_device())
            tokens = data["input_ids"].long()
            padding_mask = data["attention_mask"].long()
            if "token_type_ids" in data: types = data["token_type_ids"].long()
            else: types = None
            if "labels" in data:
                lm_labels = data["labels"].long()
                loss_mask = (data["labels"] != -100).to(torch.float)
            else:
                lm_labels = None
                loss_mask = None
            if "next_sentence_label" in data: sentence_order = data["next_sentence_label"].long()
            else: sentence_order = None
            return tokens, types, sentence_order, loss_mask, lm_labels, padding_mask
        if accelerator.state.megatron_lm_plugin.custom_get_batch_function is not None: return accelerator.state.megatron_lm_plugin.custom_get_batch_function
        if megatron_dataset_flag:
            try:
                from pretrain_bert import get_batch
                return get_batch
            except ImportError: pass
            return get_batch_megatron
        else: return get_batch_transformer
    def get_loss_func(self, accelerator, pretraining_flag, num_labels):
        def loss_func_pretrain(loss_mask, sentence_order, output_tensor):
            lm_loss_, sop_logits = output_tensor
            lm_loss_ = lm_loss_.float()
            loss_mask = loss_mask.float()
            lm_loss = torch.sum(lm_loss_.view(-1) * loss_mask.reshape(-1)) / loss_mask.sum()
            if sop_logits is not None:
                sop_loss = F.cross_entropy(sop_logits.view(-1, 2).float(), sentence_order.view(-1), ignore_index=-1)
                sop_loss = sop_loss.float()
                loss = lm_loss + sop_loss
                averaged_losses = average_losses_across_data_parallel_group([lm_loss, sop_loss])
                return loss, {"lm loss": averaged_losses[0], "sop loss": averaged_losses[1]}
            else:
                loss = lm_loss
                averaged_losses = average_losses_across_data_parallel_group([lm_loss])
                return loss, {"lm loss": averaged_losses[0]}
        def loss_func_finetune(labels, logits):
            if num_labels == 1:
                loss_fct = MSELoss()
                loss = loss_fct(logits.view(-1), labels.view(-1))
            elif self.num_labels > 1 and (labels.dtype in (torch.long, torch.int)):
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, num_labels), labels.view(-1))
            else:
                loss_fct = BCEWithLogitsLoss()
                loss = loss_fct(logits, labels)
            averaged_losses = average_losses_across_data_parallel_group([loss])
            return loss, {"loss": averaged_losses[0]}
        if accelerator.state.megatron_lm_plugin.custom_loss_function is not None: return accelerator.state.megatron_lm_plugin.custom_loss_function
        if pretraining_flag: return loss_func_pretrain
        else: return loss_func_finetune
    def get_forward_step_func(self, pretraining_flag, bert_binary_head):
        def forward_step(data_iterator, model):
            tokens, types, sentence_order, loss_mask, labels, padding_mask = self.get_batch(data_iterator)
            if not bert_binary_head: types = None
            if pretraining_flag:
                output_tensor = model(tokens, padding_mask, tokentype_ids=types, lm_labels=labels)
                return output_tensor, partial(self.loss_func, loss_mask, sentence_order)
            else:
                logits = model(tokens, padding_mask, tokentype_ids=types)
                return logits, partial(self.loss_func, labels)
        return forward_step
class GPTTrainStep(AbstractTrainStep):
    def __init__(self, accelerator, args):
        super().__init__("GPTTrainStep")
        self.get_batch = self.get_batch_func(accelerator, args.megatron_dataset_flag)
        self.loss_func = self.get_loss_func(accelerator)
        self.forward_step = self.get_forward_step_func()
        self.eod_token = args.padded_vocab_size - 1
        if args.vocab_file is not None:
            tokenizer = get_tokenizer()
            self.eod_token = tokenizer.eod
        self.reset_position_ids = args.reset_position_ids
        self.reset_attention_mask = args.reset_attention_mask
        self.eod_mask_loss = args.eod_mask_loss
        if not args.model_return_dict: self.model_output_class = None
        else:
            from transformers.modeling_outputs import CausalLMOutputWithCrossAttentions
            self.model_output_class = CausalLMOutputWithCrossAttentions
    def get_batch_func(self, accelerator, megatron_dataset_flag):
        def get_batch_megatron(data_iterator):
            keys = ["text"]
            datatype = torch.int64
            if data_iterator is not None: data = next(data_iterator)
            else: data = None
            data_b = tensor_parallel.broadcast_data(keys, data, datatype)
            tokens_ = data_b["text"].long()
            labels = tokens_[:, 1:].contiguous()
            tokens = tokens_[:, :-1].contiguous()
            attention_mask, loss_mask, position_ids = get_ltor_masks_and_position_ids(tokens, self.eod_token, self.reset_position_ids, self.reset_attention_mask, self.eod_mask_loss)
            return tokens, labels, loss_mask, attention_mask, position_ids
        def get_batch_transformer(data_iterator):
            data = next(data_iterator)
            data = {"input_ids": data["input_ids"]}
            data = send_to_device(data, torch.cuda.current_device())
            tokens_ = data["input_ids"].long()
            padding = torch.zeros((tokens_.shape[0], 1), dtype=tokens_.dtype, device=tokens_.device) + self.eod_token
            tokens_ = torch.concat([tokens_, padding], dim=1)
            labels = tokens_[:, 1:].contiguous()
            tokens = tokens_[:, :-1].contiguous()
            attention_mask, loss_mask, position_ids = get_ltor_masks_and_position_ids(tokens, self.eod_token, self.reset_position_ids, self.reset_attention_mask, True)
            return tokens, labels, loss_mask, attention_mask, position_ids
        if accelerator.state.megatron_lm_plugin.custom_get_batch_function is not None: return accelerator.state.megatron_lm_plugin.custom_get_batch_function
        if megatron_dataset_flag:
            try:
                from pretrain_gpt import get_batch
                return get_batch
            except ImportError: pass
            return get_batch_megatron
        else: return get_batch_transformer
    def get_loss_func(self, accelerator):
        args = get_args()
        def loss_func(loss_mask, output_tensor):
            if args.return_logits: losses, logits = output_tensor
            else: losses = output_tensor
            losses = losses.float()
            loss_mask = loss_mask.view(-1).float()
            if args.context_parallel_size > 1:
                loss = torch.cat([torch.sum(losses.view(-1) * loss_mask).view(1), loss_mask.sum().view(1)])
                torch.distributed.all_reduce(loss, group=mpu.get_context_parallel_group())
                loss = loss[0] / loss[1]
            else: loss = torch.sum(losses.view(-1) * loss_mask) / loss_mask.sum()
            if args.check_for_nan_in_loss_and_grad:
                global_rank = torch.distributed.get_rank()
                assert not loss.isnan(), (f"Rank {global_rank}: found NaN in local forward loss calculation. Device: {torch.cuda.current_device()}, node: {os.uname()[1]}")
            averaged_loss = average_losses_across_data_parallel_group([loss])
            output_dict = {"lm loss": averaged_loss[0]}
            if args.return_logits: output_dict.update({"logits": logits})
            return loss, output_dict
        if accelerator.state.megatron_lm_plugin.custom_loss_function is not None: return accelerator.state.megatron_lm_plugin.custom_loss_function
        return loss_func
    def get_forward_step_func(self):
        def forward_step(data_iterator, model):
            tokens, labels, loss_mask, attention_mask, position_ids = self.get_batch(data_iterator)
            output_tensor = model(tokens, position_ids, attention_mask, labels=labels)
            return output_tensor, partial(self.loss_func, loss_mask)
        return forward_step
class T5TrainStep(AbstractTrainStep):
    def __init__(self, accelerator, args):
        super().__init__("T5TrainStep")
        self.get_batch = self.get_batch_func(accelerator, args.megatron_dataset_flag)
        self.loss_func = self.get_loss_func(accelerator)
        self.forward_step = self.get_forward_step_func()
        if not args.model_return_dict: self.model_output_class = None
        else:
            from transformers.modeling_outputs import Seq2SeqLMOutput
            self.model_output_class = Seq2SeqLMOutput
    @staticmethod
    def attn_mask_postprocess(attention_mask):
        attention_mask_b1s = attention_mask.unsqueeze(1)
        attention_mask_bs1 = attention_mask.unsqueeze(2)
        attention_mask_bss = attention_mask_b1s * attention_mask_bs1
        extended_attention_mask = attention_mask_bss < 0.5
        return extended_attention_mask
    @staticmethod
    def get_decoder_mask(seq_length, device):
        attention_mask = torch.tril(torch.ones((1, seq_length, seq_length), device=device))
        attention_mask = attention_mask < 0.5
        return attention_mask
    @staticmethod
    def get_enc_dec_mask(attention_mask, dec_seq_length, device):
        batch_size, _ = attention_mask.shape
        attention_mask_b1s = attention_mask.unsqueeze(1)
        attention_mask_bs1 = torch.ones((batch_size, dec_seq_length, 1), device=device)
        attention_mask_bss = attention_mask_bs1 * attention_mask_b1s
        extended_attention_mask = attention_mask_bss < 0.5
        return extended_attention_mask
    def get_batch_func(self, accelerator, megatron_dataset_flag):
        def get_batch_megatron(data_iterator):
            keys = ["text_enc", "text_dec", "labels", "loss_mask", "enc_mask", "dec_mask", "enc_dec_mask"]
            datatype = torch.int64
            if data_iterator is not None: data = next(data_iterator)
            else: data = None
            data_b = tensor_parallel.broadcast_data(keys, data, datatype)
            tokens_enc = data_b["text_enc"].long()
            tokens_dec = data_b["text_dec"].long()
            labels = data_b["labels"].long()
            loss_mask = data_b["loss_mask"].float()
            enc_mask = data_b["enc_mask"] < 0.5
            dec_mask = data_b["dec_mask"] < 0.5
            enc_dec_mask = data_b["enc_dec_mask"] < 0.5
            return tokens_enc, tokens_dec, loss_mask, labels, enc_mask, dec_mask, enc_dec_mask
        def get_batch_transformer(data_iterator):
            data = next(data_iterator)
            data = send_to_device(data, torch.cuda.current_device())
            tokens_enc = data["input_ids"].long()
            labels = data["labels"].long()
            loss_mask = (labels != -100).to(torch.float)
            if "decoder_input_ids" in data: tokens_dec = data["decoder_input_ids"].long()
            else:
                tokens_dec = labels.new_zeros(labels.shape, device=labels.device, dtype=torch.long)
                tokens_dec[..., 1:] = labels[..., :-1].clone()
                tokens_dec[..., 0] = 0
                tokens_dec.masked_fill_(tokens_dec == -100, 0)
            enc_mask = T5TrainStep.attn_mask_postprocess(data["attention_mask"].long())
            dec_mask = T5TrainStep.get_decoder_mask(tokens_dec.shape[1], tokens_dec.device)
            enc_dec_mask = T5TrainStep.get_enc_dec_mask(data["attention_mask"].long(), tokens_dec.shape[1], tokens_dec.device)
            return tokens_enc, tokens_dec, loss_mask, labels, enc_mask, dec_mask, enc_dec_mask
        if accelerator.state.megatron_lm_plugin.custom_get_batch_function is not None: return accelerator.state.megatron_lm_plugin.custom_get_batch_function
        if megatron_dataset_flag:
            try:
                from pretrain_t5 import get_batch
                return get_batch
            except ImportError: pass
            return get_batch_megatron
        else: return get_batch_transformer
    def get_loss_func(self, accelerator):
        def loss_func(loss_mask, output_tensor):
            lm_loss_ = output_tensor.float()
            lm_loss = torch.sum(lm_loss_.view(-1) * loss_mask.reshape(-1)) / loss_mask.sum()
            loss = lm_loss
            averaged_losses = average_losses_across_data_parallel_group([lm_loss])
            return loss, {"lm loss": averaged_losses[0]}
        if accelerator.state.megatron_lm_plugin.custom_loss_function is not None: return accelerator.state.megatron_lm_plugin.custom_loss_function
        return loss_func
    def get_forward_step_func(self):
        def forward_step(data_iterator, model):
            tokens_enc, tokens_dec, loss_mask, lm_labels, enc_mask, dec_mask, enc_dec_mask = self.get_batch(data_iterator)
            output_tensor = model(tokens_enc, tokens_dec, enc_mask, dec_mask, enc_dec_mask, tokentype_ids=None, lm_labels=lm_labels)
            return output_tensor, partial(self.loss_func, loss_mask)
        return forward_step
def finish_mpu_init():
    args = get_args()
    _initialize_distributed()
    _set_random_seed(args.seed, args.data_parallel_random_init)
def initialize(accelerator, extra_args_provider=None, args_defaults={}):
    accelerator.print("Initializing Megatron-LM")
    assert torch.cuda.is_available(), "Megatron requires CUDA."
    args = parse_args(extra_args_provider, ignore_unknown_args=True)
    for key, value in args_defaults.items():
        setattr(args, key, value)
    if args.use_checkpoint_args or args_defaults.get("use_checkpoint_args", False):
        assert args.load is not None, "--use-checkpoints-args requires --load argument"
        load_args_from_checkpoint(args)
    validate_args(args)
    set_global_variables(args)
    finish_mpu_init()
    _init_autoresume()
    _compile_dependencies()
    set_jit_fusion_options()
    args = get_args()
    if getattr(args, "padded_vocab_size", None) is None: args.padded_vocab_size = _vocab_size_with_padding(args.orig_vocab_size, args)
    if args.model_type_name == "bert" and args.pretraining_flag and args.num_labels == 2: args.bert_binary_head = True
    else: args.bert_binary_head = False
    args.iteration = 0
class MegatronEngine(torch.nn.Module):
    def __init__(self, accelerator, model, optimizer, scheduler):
        super().__init__()
        self.module = model
        self.base_model = model[0]
        self.optimizer = optimizer
        self.scheduler = scheduler
        args = get_args()
        if accelerator.state.megatron_lm_plugin.custom_train_step_class is not None: self.train_step_handler = accelerator.state.megatron_lm_plugin.custom_train_step_class(args, **accelerator.state.megatron_lm_plugin.custom_train_step_kwargs)
        elif args.model_type_name == "bert": self.train_step_handler = BertTrainStep(accelerator, args)
        elif args.model_type_name == "gpt": self.train_step_handler = GPTTrainStep(accelerator, args)
        elif args.model_type_name == "t5": self.train_step_handler = T5TrainStep(accelerator, args)
        else: raise ValueError(f"Unsupported model type: {args.model_type_name}")
        self.optimizer.skipped_iter = False
        self.total_loss_dict = {}
        self.eval_total_loss_dict = {}
        self.iteration = 0
        self.report_memory_flag = True
        self.num_floating_point_operations_so_far = 0
        self.module_config = None
        if args.tensorboard_dir is not None: write_args_to_tensorboard()
    def get_module_config(self):
        args = get_args()
        config = get_model_config(self.module[0])
        config.grad_scale_func = self.optimizer.scale_loss
        if isinstance(self.module[0], LocalDDP) and args.overlap_grad_reduce:
            assert config.no_sync_func is None, ("When overlap_grad_reduce is True, config.no_sync_func must be None; a custom no_sync_func is not supported when overlapping grad-reduce")
            config.no_sync_func = [model_chunk.no_sync for model_chunk in self.module]
            if len(self.module) == 1: config.no_sync_func = config.no_sync_func[0]
            if args.delay_grad_reduce:
                config.grad_sync_func = [model_chunk.start_grad_sync for model_chunk in self.module]
                if len(self.module) == 1: config.grad_sync_func = config.grad_sync_func[0]
        if args.overlap_param_gather and args.delay_param_gather:
            config.param_sync_func = [lambda x: self.optimizer.finish_param_sync(model_index, x) for model_index in range(len(self.module))]
            if len(self.module) == 1: config.param_sync_func = config.param_sync_func[0]
        config.finalize_model_grads_func = finalize_model_grads
        return config
    def train(self):
        for model_module in self.module: model_module.train()
        if self.module_config is None: self.module_config = self.get_module_config()
        self.log_eval_results()
    def eval(self):
        for model_module in self.module: model_module.eval()
        if self.module_config is None: self.module_config = self.get_module_config()
    def get_batch_data_iterator(self, batch_data):
        args = get_args()
        data_chunks = []
        if len(batch_data) > 0:
            if args.num_micro_batches > 1:
                for i in range(0, args.num_micro_batches): data_chunks.append({k: v[i * args.micro_batch_size : (i + 1) * args.micro_batch_size] for k, v in batch_data.items()})
            else: data_chunks = [batch_data]
        if len(self.module) > 1: batch_data_iterator = ([iter(data_chunks) for _ in range(len(self.module))] if len(batch_data) > 0 else [None] * len(self.module))
        else: batch_data_iterator = iter(data_chunks) if len(batch_data) > 0 else None
        return batch_data_iterator
    def train_step(self, **batch_data):
        batch_data_iterator = self.get_batch_data_iterator(batch_data)
        loss_reduced, skipped_iter, grad_norm, num_zeros_in_grad = train_step(forward_step_func=self.train_step_handler.forward_step, data_iterator=batch_data_iterator,
        model=self.module, optimizer=self.optimizer, opt_param_scheduler=self.scheduler, config=self.module_config)
        self.optimizer.skipped_iter = skipped_iter == 1
        return loss_reduced, skipped_iter, grad_norm, num_zeros_in_grad
    def eval_step(self, **batch_data):
        args = get_args()
        batch_data_iterator = self.get_batch_data_iterator(batch_data)
        forward_backward_func = get_forward_backward_func()
        loss_dicts = forward_backward_func(forward_step_func=self.train_step_handler.forward_step, data_iterator=batch_data_iterator, model=self.module,
        num_microbatches=get_num_microbatches(), seq_length=args.seq_length, micro_batch_size=args.micro_batch_size, forward_only=True)
        if args.empty_unused_memory_level >= 1: torch.cuda.empty_cache()
        args.consumed_valid_samples += (mpu.get_data_parallel_world_size() * args.micro_batch_size * get_num_microbatches())
        if mpu.is_pipeline_last_stage(ignore_virtual=True):
            loss_reduced = {}
            for key in loss_dicts[0]:
                losses_reduced_for_key = [x[key] for x in loss_dicts]
                if len(losses_reduced_for_key[0].shape) == 0: loss_reduced[key] = sum(losses_reduced_for_key) / len(losses_reduced_for_key)
                else: loss_reduced[key] = torch.concat(losses_reduced_for_key)
            return loss_reduced
        return {}
    def forward(self, **batch_data):
        args = get_args()
        if self.module[0].training:
            loss_dict, skipped_iter, grad_norm, num_zeros_in_grad = self.train_step(**batch_data)
            self.iteration += 1
            batch_size = mpu.get_data_parallel_world_size() * args.micro_batch_size * get_num_microbatches()
            args.consumed_train_samples += batch_size
            self.num_floating_point_operations_so_far += num_floating_point_operations(args, batch_size)
            if args.tensorboard_dir is not None:
                loss_scale = self.optimizer.get_loss_scale().item()
                params_norm = None
                if args.log_params_norm: params_norm = calc_params_l2_norm(self.model)
                self.report_memory_flag = training_log(loss_dict, self.total_loss_dict, self.optimizer.param_groups[0]["lr"], self.iteration, loss_scale,
                self.report_memory_flag, skipped_iter, grad_norm, params_norm, num_zeros_in_grad)
        else:
            loss_dict = self.eval_step(**batch_data)
            if args.tensorboard_dir is not None:
                for key in loss_dict:
                    self.eval_total_loss_dict[key] = (self.eval_total_loss_dict.get(key, torch.cuda.FloatTensor([0.0])) + loss_dict[key])
                    self.eval_total_loss_dict[key + "_num_iters"] = self.eval_total_loss_dict.get(key + "_num_iters", torch.cuda.FloatTensor([0.0])) + torch.cuda.FloatTensor([1.0])
        loss = torch.tensor(0.0, device=torch.cuda.current_device())
        for key in loss_dict:
            if len(loss_dict[key].shape) == 0: loss += loss_dict[key]
        logits = None
        if "logits" in loss_dict: logits = loss_dict["logits"]
        if self.train_step_handler.model_output_class is not None: return self.train_step_handler.model_output_class(loss=loss, logits=logits)
        return loss
    def log_eval_results(self):
        args = get_args()
        if args.tensorboard_dir is None or self.iteration == 0: return
        args = get_args()
        writer = get_tensorboard_writer()
        string = f"validation loss at iteration {self.iteration} | "
        for key in self.eval_total_loss_dict:
            if key.endswith("_num_iters"): continue
            value = self.eval_total_loss_dict[key] / self.eval_total_loss_dict[key + "_num_iters"]
            string += f"{key} value: {value} | "
            ppl = math.exp(min(20, value.item()))
            if args.pretraining_flag: string += f"{key} PPL: {ppl} | "
            if writer:
                writer.add_scalar(f"{key} validation", value.item(), self.iteration)
                if args.pretraining_flag: writer.add_scalar(f"{key} validation ppl", ppl, self.iteration)
        length = len(string) + 1
        self.eval_total_loss_dict = {}
    def save_checkpoint(self, output_dir):
        self.log_eval_results()
        args = get_args()
        args.save = output_dir
        torch.distributed.barrier()
        save_checkpoint(self.iteration, self.module, self.optimizer, self.scheduler, num_floating_point_operations_so_far=self.num_floating_point_operations_so_far)
        torch.distributed.barrier()
    def load_checkpoint(self, input_dir):
        args = get_args()
        args.load = input_dir
        args.consumed_train_samples = 0
        args.consumed_valid_samples = 0
        torch.distributed.barrier()
        iteration, num_floating_point_operations_so_far = load_checkpoint(self.module, self.optimizer, self.scheduler)
        torch.distributed.barrier()
        self.iteration = iteration
        self.num_floating_point_operations_so_far = num_floating_point_operations_so_far
        if args.fp16 and self.iteration == 0: self.optimizer.reload_model_params()
    def megatron_generate(self, inputs, attention_mask=None, max_length=None, max_new_tokens=None, num_beams=None, temperature=None, top_k=None, top_p=None, length_penalty=None, **kwargs):
        args = get_args()
        if args.model_type_name != "gpt": raise NotImplementedError("Generate method is not implemented for this model")
        if args.data_parallel_size > 1: raise ValueError("Generate method requires data parallelism to be 1")
        if args.sequence_parallel: raise ValueError("Generate method requires sequence parallelism to be False")
        if args.recompute_granularity is not None: raise ValueError("Checkpoint activations cannot be set for inference")
        if args.vocab_file is None: raise ValueError("Vocab file is required for inference")
        if max_length is None and max_new_tokens is None: raise ValueError("`max_length` or `max_new_tokens` are required for inference")
        if temperature is None: temperature = 1.0
        elif not (0.0 < temperature <= 100.0): raise ValueError("temperature must be a positive number less than or equal to 100.0")
        if top_k is None: top_k = 0
        elif not (0 <= top_k <= 1000): raise ValueError("top_k must be a positive number less than or equal to 1000")
        if top_p is None: top_p = 0.0
        elif top_p > 0.0 and top_k > 0.0: raise ValueError("top_p and top_k sampling cannot be set together")
        else:
            if not (0.0 <= top_p <= 1.0): raise ValueError("top_p must be less than or equal to 1.0")
        top_p_decay = kwargs.get("top_p_decay", 0.0)
        if not (0.0 <= top_p_decay <= 1.0): raise ValueError("top_p_decay must be less than or equal to 1.0")
        top_p_bound = kwargs.get("top_p_bound", 0.0)
        if not (0.0 <= top_p_bound <= 1.0): raise ValueError("top_p_bound must be less than or equal to 1.0")
        add_BOS = kwargs.get("add_BOS", False)
        if not (isinstance(add_BOS, bool)): raise ValueError("add_BOS must be a boolean")
        beam_width = num_beams
        if beam_width is not None:
            if not isinstance(beam_width, int): raise ValueError("beam_width must be an integer")
            if beam_width < 1: raise ValueError("beam_width must be greater than 0")
            if inputs.shape[0] > 1: return "When doing beam_search, batch size must be 1"
        tokenizer = get_tokenizer()
        stop_token = kwargs.get("stop_token", tokenizer.eod)
        if stop_token is not None:
            if not isinstance(stop_token, int): raise ValueError("stop_token must be an integer")
        if length_penalty is None: length_penalty = 1.0
        sizes_list = None
        prompts_tokens_tensor = None
        prompts_length_tensor = None
        if torch.distributed.get_rank() == 0:
            if attention_mask is None: prompts_length_tensor = torch.cuda.LongTensor([inputs.shape[1]] * inputs.shape[0])
            else: prompts_length_tensor = attention_mask.sum(axis=-1).cuda()
            if max_new_tokens is None: max_new_tokens = max_length - inputs.shape[1]
            if max_new_tokens <= 0: raise ValueError("max_new_tokens must be greater than 0")
            if add_BOS:
                max_length = max_new_tokens + inputs.shape[1] + 1
                max_length = 4 * math.ceil(max_length / 4)
                max_new_tokens = max_length - (inputs.shape[1] + 1)
                padding = torch.cuda.LongTensor([[tokenizer.eod] * max_new_tokens] * inputs.shape[0])
                prompts_tokens_tensor = torch.concat([torch.unsqueeze(padding[:, 0], axis=-1), inputs.cuda(), padding], axis=-1)
            else:
                max_length = max_new_tokens + inputs.shape[1]
                max_length = 4 * math.ceil(max_length / 4)
                max_new_tokens = max_length - inputs.shape[1]
                padding = torch.cuda.LongTensor([[tokenizer.eod] * max_new_tokens] * inputs.shape[0])
                prompts_tokens_tensor = torch.concat([inputs.cuda(), padding], axis=-1)
            sizes_list = [prompts_tokens_tensor.size(0), prompts_tokens_tensor.size(1)]
        sizes_tensor = broadcast_int_list(2, int_list=sizes_list, rank=0)
        sizes = sizes_tensor.tolist()
        context_tokens_tensor = broadcast_tensor(sizes, torch.int64, tensor=prompts_tokens_tensor, rank=0)
        context_length_tensor = broadcast_tensor(sizes[0], torch.int64, tensor=prompts_length_tensor, rank=0)
        random_seed = kwargs.get("random_seed", 0)
        torch.random.manual_seed(random_seed)
        unwrapped_model = unwrap_model(self.base_model, (torchDDP, LocalDDP, Float16Module))
        if beam_width is not None: tokens, _ = beam_search_and_return_on_first_stage(unwrapped_model, context_tokens_tensor, context_length_tensor, beam_width, stop_token=stop_token, num_return_gen=1, length_penalty=length_penalty)
        else: tokens, _, _ = generate_tokens_probs_and_return_on_first_stage(unwrapped_model, context_tokens_tensor, context_length_tensor, return_output_log_probs=False,
        top_k=top_k, top_p=top_p, top_p_decay=top_p_decay, top_p_bound=top_p_bound, temperature=temperature, use_eod_token_for_early_termination=True)
        return tokens
def avg_losses_across_data_parallel_group(losses): return average_losses_across_data_parallel_group(losses)
def gather_across_data_parallel_groups(tensor):
    def _gpu_gather_one(tensor):
        if tensor.ndim == 0: tensor = tensor.clone()[None]
        output_tensors = [torch.empty_like(tensor) for _ in range(torch.distributed.get_world_size(group=mpu.get_data_parallel_group()))]
        torch.distributed.all_gather(output_tensors, tensor, group=mpu.get_data_parallel_group())
        return torch.cat(output_tensors, dim=0)
    return recursively_apply(_gpu_gather_one, tensor, error_on_other_type=True)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import collections
import os
import platform
import re
import socket
from contextlib import contextmanager
from functools import partial, reduce
from types import MethodType
from typing import OrderedDict
import torch
from packaging.version import Version
from safetensors.torch import save_file as safe_save_file
from ..commands.config.default import write_basic_config
from ..logging import get_logger
from ..state import PartialState
from .constants import FSDP_PYTORCH_VERSION
from .dataclasses import DistributedType
from .imports import is_deepspeed_available, is_torch_distributed_available, is_torch_xla_available
from .modeling import id_tensor_storage
from .transformer_engine import convert_model
from .versions import is_torch_version
logger = get_logger(__name__)
if is_torch_xla_available(): import torch_xla.core.xla_model as xm
def is_compiled_module(module):
    if is_torch_version("<", "2.0.0") or not hasattr(torch, "_dynamo"): return False
    return isinstance(module, torch._dynamo.eval_frame.OptimizedModule)
def extract_model_from_parallel(model, keep_fp32_wrapper: bool = True, recursive: bool = False):
    options = (torch.nn.parallel.DistributedDataParallel, torch.nn.DataParallel)
    is_compiled = is_compiled_module(model)
    if is_compiled:
        compiled_model = model
        model = model._orig_mod
    if is_deepspeed_available():
        from deepspeed import DeepSpeedEngine
        options += (DeepSpeedEngine,)
    if is_torch_version(">=", FSDP_PYTORCH_VERSION) and is_torch_distributed_available():
        from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
        options += (FSDP,)
    while isinstance(model, options): model = model.module
    if recursive:
        def _recursive_unwrap(module):
            if hasattr(module, "module"): unwrapped_module = _recursive_unwrap(module.module)
            else: unwrapped_module = module
            for name, child in unwrapped_module.named_children(): setattr(unwrapped_module, name, _recursive_unwrap(child))
            return unwrapped_module
        model = _recursive_unwrap(model)
    if not keep_fp32_wrapper:
        forward = model.forward
        original_forward = model.__dict__.pop("_original_forward", None)
        if original_forward is not None:
            while hasattr(forward, "__wrapped__"):
                forward = forward.__wrapped__
                if forward == original_forward: break
            model.forward = MethodType(forward, model)
        if getattr(model, "_converted_to_transformer_engine", False): convert_model(model, to_transformer_engine=False)
    if is_compiled:
        compiled_model._orig_mod = model
        model = compiled_model
    return model
def wait_for_everyone(): PartialState().wait_for_everyone()
def clean_state_dict_for_safetensors(state_dict: dict):
    ptrs = collections.defaultdict(list)
    for name, tensor in state_dict.items():
        if not isinstance(tensor, str): ptrs[id_tensor_storage(tensor)].append(name)
    shared_ptrs = {ptr: names for ptr, names in ptrs.items() if len(names) > 1}
    warn_names = set()
    for names in shared_ptrs.values():
        found_names = [name for name in names if name in state_dict]
        warn_names.update(found_names[1:])
        for name in found_names[1:]: del state_dict[name]
    state_dict = {k: v.contiguous() if isinstance(v, torch.Tensor) else v for k, v in state_dict.items()}
    return state_dict
def save(obj, f, save_on_each_node: bool = False, safe_serialization: bool = False):
    if PartialState().distributed_type == DistributedType.XLA: obj = xm._maybe_convert_to_cpu(obj)
    if safe_serialization:
        save_func = partial(safe_save_file, metadata={"format": "pt"})
        if isinstance(obj, OrderedDict): obj = clean_state_dict_for_safetensors(obj)
    else: save_func = torch.save
    if PartialState().is_main_process and not save_on_each_node: save_func(obj, f)
    elif PartialState().is_local_main_process and save_on_each_node: save_func(obj, f)
@contextmanager
def clear_environment():
    _old_os_environ = os.environ.copy()
    os.environ.clear()
    try: yield
    finally:
        os.environ.clear()
        os.environ.update(_old_os_environ)
@contextmanager
def patch_environment(**kwargs):
    existing_vars = {}
    for key, value in kwargs.items():
        key = key.upper()
        if key in os.environ: existing_vars[key] = os.environ[key]
        os.environ[key] = str(value)
    try: yield
    finally:
        for key in kwargs:
            key = key.upper()
            if key in existing_vars: os.environ[key] = existing_vars[key]
            else: os.environ.pop(key, None)
def get_pretty_name(obj):
    if not hasattr(obj, "__qualname__") and not hasattr(obj, "__name__"): obj = getattr(obj, "__class__", obj)
    if hasattr(obj, "__qualname__"): return obj.__qualname__
    if hasattr(obj, "__name__"): return obj.__name__
    return str(obj)
def merge_dicts(source, destination):
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            merge_dicts(value, node)
        else: destination[key] = value
    return destination
def is_port_in_use(port: int = None) -> bool:
    if port is None: port = 29500
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s: return s.connect_ex(("localhost", port)) == 0
def convert_bytes(size):
    for x in ["bytes", "KB", "MB", "GB", "TB"]:
        if size < 1024.0: return f"{round(size, 2)} {x}"
        size /= 1024.0
    return f"{round(size, 2)} PB"
def check_os_kernel():
    info = platform.uname()
    system = info.system
    if system != "Linux": return
    _, version, *_ = re.split(r"(\d+\.\d+\.\d+)", info.release)
    min_version = "5.5.0"
def recursive_getattr(obj, attr: str):
    def _getattr(obj, attr): return getattr(obj, attr)
    return reduce(_getattr, [obj] + attr.split("."))
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import random
from typing import List, Optional, Union
import numpy as np
import torch
from ..state import AcceleratorState
from .constants import CUDA_DISTRIBUTED_TYPES
from .dataclasses import DistributedType, RNGType
from .imports import is_mlu_available, is_musa_available, is_npu_available, is_torch_xla_available, is_xpu_available
if is_torch_xla_available(): import torch_xla.core.xla_model as xm
def set_seed(seed: int, device_specific: bool = False, deterministic: bool = False):
    if device_specific: seed += AcceleratorState().process_index
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if is_xpu_available(): torch.xpu.manual_seed_all(seed)
    elif is_npu_available(): torch.npu.manual_seed_all(seed)
    elif is_mlu_available(): torch.mlu.manual_seed_all(seed)
    elif is_musa_available(): torch.musa.manual_seed_all(seed)
    else: torch.cuda.manual_seed_all(seed)
    if is_torch_xla_available(): xm.set_rng_state(seed)
    if deterministic: torch.use_deterministic_algorithms(True)
def synchronize_rng_state(rng_type: Optional[RNGType] = None, generator: Optional[torch.Generator] = None):
    if rng_type == RNGType.TORCH: rng_state = torch.get_rng_state()
    elif rng_type == RNGType.CUDA: rng_state = torch.cuda.get_rng_state()
    elif rng_type == RNGType.XLA:
        assert is_torch_xla_available(), "Can't synchronize XLA seeds as torch_xla is unavailable."
        rng_state = torch.tensor(xm.get_rng_state())
    elif rng_type == RNGType.NPU:
        assert is_npu_available(), "Can't synchronize NPU seeds on an environment without NPUs."
        rng_state = torch.npu.get_rng_state()
    elif rng_type == RNGType.MLU:
        assert is_mlu_available(), "Can't synchronize MLU seeds on an environment without MLUs."
        rng_state = torch.mlu.get_rng_state()
    elif rng_type == RNGType.MUSA:
        assert is_musa_available(), "Can't synchronize MUSA seeds on an environment without MUSAs."
        rng_state = torch.musa.get_rng_state()
    elif rng_type == RNGType.XPU:
        assert is_xpu_available(), "Can't synchronize XPU seeds on an environment without XPUs."
        rng_state = torch.xpu.get_rng_state()
    elif rng_type == RNGType.GENERATOR:
        assert generator is not None, "Need a generator to synchronize its seed."
        rng_state = generator.get_state()
    state = AcceleratorState()
    if state.distributed_type == DistributedType.XLA:
        rng_state = rng_state.to(xm.xla_device())
        xm.collective_broadcast([rng_state])
        xm.mark_step()
        rng_state = rng_state.cpu()
    elif (state.distributed_type in CUDA_DISTRIBUTED_TYPES or state.distributed_type == DistributedType.MULTI_MLU or state.distributed_type == DistributedType.MULTI_MUSA
    or state.distributed_type == DistributedType.MULTI_NPU or state.distributed_type == DistributedType.MULTI_XPU):
        rng_state = rng_state.to(state.device)
        torch.distributed.broadcast(rng_state, 0)
        rng_state = rng_state.cpu()
    elif state.distributed_type == DistributedType.MULTI_CPU: torch.distributed.broadcast(rng_state, 0)
    if rng_type == RNGType.TORCH: torch.set_rng_state(rng_state)
    elif rng_type == RNGType.CUDA: torch.cuda.set_rng_state(rng_state)
    elif rng_type == RNGType.NPU: torch.npu.set_rng_state(rng_state)
    elif rng_type == RNGType.MLU: torch.mlu.set_rng_state(rng_state)
    elif rng_type == RNGType.MUSA: torch.musa.set_rng_state(rng_state)
    elif rng_type == RNGType.XPU: torch.xpu.set_rng_state(rng_state)
    elif rng_type == RNGType.XLA: xm.set_rng_state(rng_state.item())
    elif rng_type == RNGType.GENERATOR: generator.set_state(rng_state)
def synchronize_rng_states(rng_types: List[Union[str, RNGType]], generator: Optional[torch.Generator] = None):
    for rng_type in rng_types: synchronize_rng_state(RNGType(rng_type), generator=generator)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
